# Copyright (c) 2016 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import re

from lxml import builder
from lxml import etree as ET
from oslo_concurrency import processutils
from oslo_log import log
import six

from manila.common import constants as const
from manila import exception
from manila.i18n import _
from manila.share.drivers.dell_emc.common.enas import connector
from manila.share.drivers.dell_emc.common.enas import constants
from manila.share.drivers.dell_emc.common.enas import utils as vmax_utils
from manila.share.drivers.dell_emc.common.enas import xml_api_parser as parser
from manila import utils

LOG = log.getLogger(__name__)


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class StorageObjectManager(object):
    def __init__(self, configuration):
        self.context = {}

        self.connectors = {}
        self.connectors['XML'] = connector.XMLAPIConnector(configuration)
        self.connectors['SSH'] = connector.SSHConnector(configuration)

        elt_maker = builder.ElementMaker(nsmap={None: constants.XML_NAMESPACE})
        xml_parser = parser.XMLAPIParser()

        obj_types = StorageObject.__subclasses__()  # pylint: disable=no-member
        for item in obj_types:
            key = item.__name__
            self.context[key] = eval(key)(self.connectors,
                                          elt_maker,
                                          xml_parser,
                                          self)

    def getStorageContext(self, type):
        if type in self.context:
            return self.context[type]
        else:
            message = (_("Invalid storage object type %s.") % type)
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)


class StorageObject(object):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        self.conn = conn
        self.elt_maker = elt_maker
        self.xml_parser = xml_parser
        self.manager = manager
        self.xml_retry = False
        self.ssh_retry_patterns = [
            (
                constants.SSH_DEFAULT_RETRY_PATTERN,
                exception.EMCVmaxLockRequiredException()
            ),
        ]

    def _translate_response(self, response):
        """Translate different status to ok/error status."""
        if (constants.STATUS_OK == response['maxSeverity'] or
                constants.STATUS_ERROR == response['maxSeverity']):
            return

        old_Severity = response['maxSeverity']
        if response['maxSeverity'] in (constants.STATUS_DEBUG,
                                       constants.STATUS_INFO):
            response['maxSeverity'] = constants.STATUS_OK

            LOG.warning("Translated status from %(old)s to %(new)s. "
                        "Message: %(info)s.",
                        {'old': old_Severity,
                         'new': response['maxSeverity'],
                         'info': response})

    def _response_validation(self, response, error_code):
        """Validates whether a response includes a certain error code."""
        msg_codes = self._get_problem_message_codes(response['problems'])

        for code in msg_codes:
            if code == error_code:
                return True

        return False

    def _get_problem_message_codes(self, problems):
        message_codes = []
        for problem in problems:
            if 'messageCode' in problem:
                message_codes.append(problem['messageCode'])

        return message_codes

    def _get_problem_messages(self, problems):
        messages = []
        for problem in problems:
            if 'message' in problem:
                messages.append(problem['message'])

        return messages

    def _get_problem_diags(self, problems):
        diags = []

        for problem in problems:
            if 'Diagnostics' in problem:
                diags.append(problem['Diagnostics'])

        return diags

    def _build_query_package(self, body):
        return self.elt_maker.RequestPacket(
            self.elt_maker.Request(
                self.elt_maker.Query(body)
            )
        )

    def _build_task_package(self, body):
        return self.elt_maker.RequestPacket(
            self.elt_maker.Request(
                self.elt_maker.StartTask(body, timeout='300')
            )
        )

    @utils.retry(exception.EMCVmaxLockRequiredException)
    def _send_request(self, req):
        req_xml = constants.XML_HEADER + ET.tostring(req).decode('utf-8')

        rsp_xml = self.conn['XML'].request(str(req_xml))

        response = self.xml_parser.parse(rsp_xml)

        self._translate_response(response)

        if (response['maxSeverity'] != constants.STATUS_OK and
                self._response_validation(response,
                                          constants.MSG_CODE_RETRY)):
            raise exception.EMCVmaxLockRequiredException

        return response

    @utils.retry(exception.EMCVmaxLockRequiredException)
    def _execute_cmd(self, cmd, retry_patterns=None, check_exit_code=False):
        """Execute NAS command via SSH.

        :param retry_patterns: list of tuples,where each tuple contains a reg
            expression and an exception.
        :param check_exit_code: Boolean. Raise
            processutils.ProcessExecutionError if the command failed to
            execute and this parameter is set to True.
        """
        if retry_patterns is None:
            retry_patterns = self.ssh_retry_patterns

        try:
            out, err = self.conn['SSH'].run_ssh(cmd, check_exit_code)
        except processutils.ProcessExecutionError as e:
            for pattern in retry_patterns:
                if re.search(pattern[0], e.stdout):
                    raise pattern[1]

            raise

        return out, err

    def _copy_properties(self, source, target, property_map, deep_copy=True):
        for prop in property_map:
            if isinstance(prop, tuple):
                target_key, src_key = prop
            else:
                target_key = src_key = prop

            if src_key in source:
                if deep_copy and isinstance(source[src_key], list):
                    target[target_key] = copy.deepcopy(source[src_key])
                else:
                    target[target_key] = source[src_key]
            else:
                target[target_key] = None

    def _get_mover_id(self, mover_name, is_vdm):
        if is_vdm:
            return self.get_context('VDM').get_id(mover_name)
        else:
            return self.get_context('Mover').get_id(mover_name,
                                                    self.xml_retry)

    def get_context(self, type):
        return self.manager.getStorageContext(type)


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class FileSystem(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(FileSystem, self).__init__(conn, elt_maker, xml_parser, manager)
        self.filesystem_map = {}

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def create(self, name, size, pool_name, mover_name, is_vdm=True):
        pool_id = self.get_context('StoragePool').get_id(pool_name)

        mover_id = self._get_mover_id(mover_name, is_vdm)
        if is_vdm:
            mover = self.elt_maker.Vdm(vdm=mover_id)
        else:
            mover = self.elt_maker.Mover(mover=mover_id)

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_task_package(
            self.elt_maker.NewFileSystem(
                mover,
                self.elt_maker.StoragePool(
                    pool=pool_id,
                    size=six.text_type(size),
                    mayContainSlices='true'
                ),
                name=name
            )
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif self._response_validation(
                response, constants.MSG_FILESYSTEM_EXIST):
            LOG.warning("File system %s already exists. "
                        "Skip the creation.", name)
            return
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to create file system %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    def get(self, name):
        if name not in self.filesystem_map:
            request = self._build_query_package(
                self.elt_maker.FileSystemQueryParams(
                    self.elt_maker.AspectSelection(
                        fileSystems='true',
                        fileSystemCapacityInfos='true'
                    ),
                    self.elt_maker.Alias(name=name)
                )
            )

            response = self._send_request(request)

            if constants.STATUS_OK != response['maxSeverity']:
                if self._is_filesystem_nonexistent(response):
                    return constants.STATUS_NOT_FOUND, response['problems']
                else:
                    return response['maxSeverity'], response['problems']

            if not response['objects']:
                return constants.STATUS_NOT_FOUND, response['problems']

            src = response['objects'][0]
            filesystem = {}
            property_map = (
                'name',
                ('pools_id', 'storagePools'),
                ('volume_id', 'volume'),
                ('size', 'volumeSize'),
                ('id', 'fileSystem'),
                'type',
                'dataServicePolicies',
            )

            self._copy_properties(src, filesystem, property_map)

            self.filesystem_map[name] = filesystem

        return constants.STATUS_OK, self.filesystem_map[name]

    def delete(self, name):
        status, out = self.get(name)
        if constants.STATUS_NOT_FOUND == status:
            LOG.warning("File system %s not found. Skip the deletion.",
                        name)
            return
        elif constants.STATUS_OK != status:
            message = (_("Failed to get file system by name %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': out})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        enas_id = self.filesystem_map[name]['id']

        request = self._build_task_package(
            self.elt_maker.DeleteFileSystem(fileSystem=enas_id)
        )

        response = self._send_request(request)

        if constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to delete file system %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        self.filesystem_map.pop(name)

    def extend(self, name, pool_name, new_size):
        status, out = self.get(name)
        if constants.STATUS_OK != status:
            message = (_("Failed to get file system by name %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': out})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        enas_id = out['id']
        size = int(out['size'])
        if new_size < size:
            message = (_("Failed to extend file system %(name)s because new "
                         "size %(new_size)d is smaller than old size "
                         "%(size)d.") %
                       {'name': name, 'new_size': new_size, 'size': size})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)
        elif new_size == size:
            return

        pool_id = self.get_context('StoragePool').get_id(pool_name)

        request = self._build_task_package(
            self.elt_maker.ExtendFileSystem(
                self.elt_maker.StoragePool(
                    pool=pool_id,
                    size=six.text_type(new_size - size)
                ),
                fileSystem=enas_id,
            )
        )

        response = self._send_request(request)

        if constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to extend file system %(name)s to new size "
                         "%(new_size)d. Reason: %(err)s.") %
                       {'name': name,
                        'new_size': new_size,
                        'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    def get_id(self, name):
        status, out = self.get(name)
        if constants.STATUS_OK != status:
            message = (_("Failed to get file system by name %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': out})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        return self.filesystem_map[name]['id']

    def _is_filesystem_nonexistent(self, response):
        """Translate different status to ok/error status."""
        msg_codes = self._get_problem_message_codes(response['problems'])
        diags = self._get_problem_diags(response['problems'])

        for code, diagnose in zip(msg_codes, diags):
            if (code == constants.MSG_FILESYSTEM_NOT_FOUND and
                    diagnose.find('File system not found.') != -1):
                return True

        return False

    def create_from_snapshot(self, name, snap_name, source_fs_name, pool_name,
                             mover_name, connect_id):
        create_fs_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_fs',
            '-name', name,
            '-type', 'uxfs',
            '-create',
            'samesize=' + source_fs_name,
            'pool=%s' % pool_name,
            'storage=SINGLE',
            'worm=off',
            '-thin', 'no',
            '-option', 'slice=y',
        ]

        self._execute_cmd(create_fs_cmd)

        ro_mount_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/server_mount', mover_name,
            '-option', 'ro',
            name,
            '/%s' % name,
        ]
        self._execute_cmd(ro_mount_cmd)

        session_name = name + ':' + snap_name
        copy_ckpt_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_copy',
            '-name', session_name[0:63],
            '-source', '-ckpt', snap_name,
            '-destination', '-fs', name,
            '-interconnect',
            'id=%s' % connect_id,
            '-overwrite_destination',
            '-full_copy',
        ]

        try:
            self._execute_cmd(copy_ckpt_cmd, check_exit_code=True)
        except processutils.ProcessExecutionError as expt:
            LOG.error("Failed to copy content from snapshot %(snap)s to "
                      "file system %(filesystem)s. Reason: %(err)s.",
                      {'snap': snap_name,
                       'filesystem': name,
                       'err': expt})

        # When an error happens during nas_copy, we need to continue
        # deleting the checkpoint of the target file system if it exists.
        query_fs_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_fs',
            '-info', name,
        ]
        out, err = self._execute_cmd(query_fs_cmd)
        re_ckpts = r'ckpts\s*=\s*(.*)\s*'
        m = re.search(re_ckpts, out)
        if m is not None:
            ckpts = m.group(1)
            for ckpt in re.split(',', ckpts):
                umount_ckpt_cmd = [
                    'env', 'NAS_DB=/nas',
                    '/nas/bin/server_umount', mover_name,
                    '-perm', ckpt,
                ]
                self._execute_cmd(umount_ckpt_cmd)
                delete_ckpt_cmd = [
                    'env', 'NAS_DB=/nas', '/nas/bin/nas_fs',
                    '-delete', ckpt,
                    '-Force',
                ]
                self._execute_cmd(delete_ckpt_cmd)

        rw_mount_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/server_mount', mover_name,
            '-option', 'rw',
            name,
            '/%s' % name,
        ]
        self._execute_cmd(rw_mount_cmd)


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class StoragePool(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(StoragePool, self).__init__(conn, elt_maker, xml_parser, manager)
        self.pool_map = {}

    def get(self, name, force=False):
        if name not in self.pool_map or force:
            status, out = self.get_all()
            if constants.STATUS_OK != status:
                return status, out

        if name not in self.pool_map:
            return constants.STATUS_NOT_FOUND, None

        return constants.STATUS_OK, self.pool_map[name]

    def get_all(self):
        self.pool_map.clear()

        request = self._build_query_package(
            self.elt_maker.StoragePoolQueryParams()
        )

        response = self._send_request(request)

        if constants.STATUS_OK != response['maxSeverity']:
            return response['maxSeverity'], response['problems']

        if not response['objects']:
            return constants.STATUS_NOT_FOUND, response['problems']

        for item in response['objects']:
            pool = {}
            property_map = (
                'name',
                ('movers_id', 'movers'),
                ('total_size', 'autoSize'),
                ('used_size', 'usedSize'),
                'diskType',
                'dataServicePolicies',
                ('id', 'pool'),
            )
            self._copy_properties(item, pool, property_map)
            self.pool_map[item['name']] = pool

        return constants.STATUS_OK, self.pool_map

    def get_id(self, name):
        status, out = self.get(name)

        if constants.STATUS_OK != status:
            message = (_("Failed to get storage pool by name %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': out})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        return out['id']


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class MountPoint(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(MountPoint, self).__init__(conn, elt_maker, xml_parser, manager)

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def create(self, mount_path, fs_name, mover_name, is_vdm=True):
        fs_id = self.get_context('FileSystem').get_id(fs_name)

        mover_id = self._get_mover_id(mover_name, is_vdm)

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_task_package(
            self.elt_maker.NewMount(
                self.elt_maker.MoverOrVdm(
                    mover=mover_id,
                    moverIdIsVdm='true' if is_vdm else 'false',
                ),
                fileSystem=fs_id,
                path=mount_path
            )
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif self._is_mount_point_already_existent(response):
            LOG.warning("Mount Point %(mount)s already exists. "
                        "Skip the creation.", {'mount': mount_path})
            return
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_('Failed to create Mount Point %(mount)s for '
                         'file system %(fs_name)s. Reason: %(err)s.') %
                       {'mount': mount_path,
                        'fs_name': fs_name,
                        'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def get(self, mover_name, is_vdm=True):
        mover_id = self._get_mover_id(mover_name, is_vdm)

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_query_package(
            self.elt_maker.MountQueryParams(
                self.elt_maker.MoverOrVdm(
                    mover=mover_id,
                    moverIdIsVdm='true' if is_vdm else 'false'
                )
            )
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif constants.STATUS_OK != response['maxSeverity']:
            return response['maxSeverity'], response['objects']

        if not response['objects']:
            return constants.STATUS_NOT_FOUND, None
        else:
            return constants.STATUS_OK, response['objects']

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def delete(self, mount_path, mover_name, is_vdm=True):
        mover_id = self._get_mover_id(mover_name, is_vdm)

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_task_package(
            self.elt_maker.DeleteMount(
                mover=mover_id,
                moverIdIsVdm='true' if is_vdm else 'false',
                path=mount_path
            )
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif self._is_mount_point_nonexistent(response):
            LOG.warning('Mount point %(mount)s on mover %(mover_name)s '
                        'not found.',
                        {'mount': mount_path, 'mover_name': mover_name})

            return
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_('Failed to delete mount point %(mount)s on mover '
                         '%(mover_name)s. Reason: %(err)s.') %
                       {'mount': mount_path,
                        'mover_name': mover_name,
                        'err': response})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    def _is_mount_point_nonexistent(self, response):
        """Translate different status to ok/error status."""
        msg_codes = self._get_problem_message_codes(response['problems'])
        message = self._get_problem_messages(response['problems'])

        for code, msg in zip(msg_codes, message):
            if ((code == constants.MSG_GENERAL_ERROR and msg.find(
                    'No such path or invalid operation') != -1) or
                    code == constants.MSG_INVALID_VDM_ID or
                    code == constants.MSG_INVALID_MOVER_ID):
                return True

        return False

    def _is_mount_point_already_existent(self, response):
        """Translate different status to ok/error status."""
        msg_codes = self._get_problem_message_codes(response['problems'])
        message = self._get_problem_messages(response['problems'])

        for code, msg in zip(msg_codes, message):
            if ((code == constants.MSG_GENERAL_ERROR and msg.find(
                    'Mount already exists') != -1)):
                return True

        return False


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class Mover(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(Mover, self).__init__(conn, elt_maker, xml_parser, manager)
        self.mover_map = {}
        self.mover_ref_map = {}

    def get_ref(self, name, force=False):
        if name not in self.mover_ref_map or force:
            self.mover_ref_map.clear()

            request = self._build_query_package(
                self.elt_maker.MoverQueryParams(
                    self.elt_maker.AspectSelection(movers='true')
                )
            )

            response = self._send_request(request)

            if constants.STATUS_ERROR == response['maxSeverity']:
                return response['maxSeverity'], response['problems']

            for item in response['objects']:
                mover = {}
                property_map = ('name', ('id', 'mover'))
                self._copy_properties(item, mover, property_map)
                if mover:
                    self.mover_ref_map[mover['name']] = mover

        if (name not in self.mover_ref_map or
                self.mover_ref_map[name]['id'] == ''):
            return constants.STATUS_NOT_FOUND, None

        return constants.STATUS_OK, self.mover_ref_map[name]

    def get(self, name, force=False):
        if name not in self.mover_map or force:
            if name in self.mover_ref_map and not force:
                mover_id = self.mover_ref_map[name]['id']
            else:
                mover_id = self.get_id(name, force)

            if name in self.mover_map:
                self.mover_map.pop(name)

            request = self._build_query_package(
                self.elt_maker.MoverQueryParams(
                    self.elt_maker.AspectSelection(
                        moverDeduplicationSettings='true',
                        moverDnsDomains='true',
                        moverInterfaces='true',
                        moverNetworkDevices='true',
                        moverNisDomains='true',
                        moverRoutes='true',
                        movers='true',
                        moverStatuses='true'
                    ),
                    mover=mover_id
                )
            )

            response = self._send_request(request)
            if constants.STATUS_ERROR == response['maxSeverity']:
                return response['maxSeverity'], response['problems']

            if not response['objects']:
                return constants.STATUS_NOT_FOUND, response['problems']

            mover = {}
            src = response['objects'][0]
            property_map = (
                'name',
                ('id', 'mover'),
                ('Status', 'maxSeverity'),
                'version',
                'uptime',
                'role',
                ('interfaces', 'MoverInterface'),
                ('devices', 'LogicalNetworkDevice'),
                ('dns_domain', 'MoverDnsDomain'),
            )

            self._copy_properties(src, mover, property_map)

            internal_devices = []
            if mover['interfaces']:
                for interface in mover['interfaces']:
                    if self._is_internal_device(interface['device']):
                        internal_devices.append(interface)

                mover['interfaces'] = [var for var in mover['interfaces'] if
                                       var not in internal_devices]

            self.mover_map[name] = mover

        return constants.STATUS_OK, self.mover_map[name]

    def get_id(self, name, force=False):
        status, mover_ref = self.get_ref(name, force)
        if constants.STATUS_OK != status:
            message = (_("Failed to get mover by name %(name)s.") %
                       {'name': name})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        return mover_ref['id']

    def _is_internal_device(self, device):
        for device_type in ('mge', 'fxg', 'tks', 'fsn'):
            if device.find(device_type) == 0:
                return True
        return False

    def get_interconnect_id(self, source, destination):
        header = [
            'id',
            'name',
            'source_server',
            'destination_system',
            'destination_server',
        ]

        conn_id = None

        command_nas_cel = [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_cel',
            '-interconnect', '-l',
        ]
        out, err = self._execute_cmd(command_nas_cel)

        lines = out.strip().split('\n')
        for line in lines:
            if line.strip().split() == header:
                LOG.info('Found the header of the command '
                         '/nas/bin/nas_cel -interconnect -l.')
            else:
                interconn = line.strip().split()
                if interconn[2] == source and interconn[4] == destination:
                    conn_id = interconn[0]

        return conn_id

    def get_physical_devices(self, mover_name):

        physical_network_devices = []

        cmd_sysconfig = [
            'env', 'NAS_DB=/nas', '/nas/bin/server_sysconfig', mover_name,
            '-pci'
        ]

        out, err = self._execute_cmd(cmd_sysconfig)

        re_pattern = ('0:\s*(?P<name>\S+)\s*IRQ:\s*(?P<irq>\d+)\n'
                      '.*\n'
                      '\s*Link:\s*(?P<link>[A-Za-z]+)')

        for device in re.finditer(re_pattern, out):
            if 'Up' in device.group('link'):
                physical_network_devices.append(device.group('name'))

        return physical_network_devices


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class VDM(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(VDM, self).__init__(conn, elt_maker, xml_parser, manager)
        self.vdm_map = {}

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def create(self, name, mover_name):
        mover_id = self._get_mover_id(mover_name, False)

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_task_package(
            self.elt_maker.NewVdm(mover=mover_id, name=name)
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif self._response_validation(response, constants.MSG_VDM_EXIST):
            LOG.warning("VDM %(name)s already exists. Skip the creation.",
                        {'name': name})
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to create VDM %(name)s on mover "
                         "%(mover_name)s. Reason: %(err)s.") %
                       {'name': name,
                        'mover_name': mover_name,
                        'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    def get(self, name):
        if name not in self.vdm_map:
            request = self._build_query_package(
                self.elt_maker.VdmQueryParams()
            )

            response = self._send_request(request)

            if constants.STATUS_OK != response['maxSeverity']:
                return response['maxSeverity'], response['problems']
            elif not response['objects']:
                return constants.STATUS_NOT_FOUND, response['problems']

            for item in response['objects']:
                vdm = {}
                property_map = (
                    'name',
                    ('id', 'vdm'),
                    'state',
                    ('host_mover_id', 'mover'),
                    ('interfaces', 'Interfaces'),
                )
                self._copy_properties(item, vdm, property_map)
                self.vdm_map[item['name']] = vdm

        if name not in self.vdm_map:
            return constants.STATUS_NOT_FOUND, None

        return constants.STATUS_OK, self.vdm_map[name]

    def delete(self, name):
        status, out = self.get(name)
        if constants.STATUS_NOT_FOUND == status:
            LOG.warning("VDM %s not found. Skip the deletion.",
                        name)
            return
        elif constants.STATUS_OK != status:
            message = (_("Failed to get VDM by name %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': out})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        vdm_id = self.vdm_map[name]['id']

        request = self._build_task_package(
            self.elt_maker.DeleteVdm(vdm=vdm_id)
        )

        response = self._send_request(request)

        if constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to delete VDM %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        self.vdm_map.pop(name)

    def get_id(self, name):
        status, vdm = self.get(name)
        if constants.STATUS_OK != status:
            message = (_("Failed to get VDM by name %(name)s.") %
                       {'name': name})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        return vdm['id']

    def attach_nfs_interface(self, vdm_name, if_name):

        command_attach_nfs_interface = [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_server',
            '-vdm', vdm_name,
            '-attach', if_name,
        ]

        self._execute_cmd(command_attach_nfs_interface)

    def detach_nfs_interface(self, vdm_name, if_name):

        command_detach_nfs_interface = [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_server',
            '-vdm', vdm_name,
            '-detach', if_name,
        ]

        try:
            self._execute_cmd(command_detach_nfs_interface,
                              check_exit_code=True)
        except processutils.ProcessExecutionError:
            interfaces = self.get_interfaces(vdm_name)
            if if_name not in interfaces['nfs']:
                LOG.debug("Failed to detach interface %(interface)s "
                          "from mover %(mover_name)s.",
                          {'interface': if_name, 'mover_name': vdm_name})
            else:
                message = (_("Failed to detach interface %(interface)s "
                             "from mover %(mover_name)s.") %
                           {'interface': if_name, 'mover_name': vdm_name})
                LOG.exception(message)
                raise exception.EMCVmaxXMLAPIError(err=message)

    def get_interfaces(self, vdm_name):
        interfaces = {
            'cifs': [],
            'nfs': [],
        }

        re_pattern = ('Interfaces to services mapping:'
                      '\s*(?P<interfaces>(\s*interface=.*)*)')

        command_get_interfaces = [
            'env', 'NAS_DB=/nas', '/nas/bin/nas_server',
            '-i',
            '-vdm', vdm_name,
        ]

        out, err = self._execute_cmd(command_get_interfaces)

        m = re.search(re_pattern, out)
        if m:
            if_list = m.group('interfaces').split('\n')
            for i in if_list:
                m_if = re.search('\s*interface=(?P<if>.*)\s*:'
                                 '\s*(?P<type>.*)\s*', i)
                if m_if:
                    if_name = m_if.group('if').strip()
                    if 'cifs' == m_if.group('type') and if_name != '':
                        interfaces['cifs'].append(if_name)
                    elif (m_if.group('type') in ('vdm', 'nfs')
                          and if_name != ''):
                        interfaces['nfs'].append(if_name)

        return interfaces


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class Snapshot(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(Snapshot, self).__init__(conn, elt_maker, xml_parser, manager)
        self.snap_map = {}

    def create(self, name, fs_name, pool_id, ckpt_size=None):
        fs_id = self.get_context('FileSystem').get_id(fs_name)

        if ckpt_size:
            elt_pool = self.elt_maker.StoragePool(
                pool=pool_id,
                size=six.text_type(ckpt_size)
            )
        else:
            elt_pool = self.elt_maker.StoragePool(pool=pool_id)

        new_ckpt = self.elt_maker.NewCheckpoint(
            self.elt_maker.SpaceAllocationMethod(
                elt_pool
            ),
            checkpointOf=fs_id,
            name=name
        )

        request = self._build_task_package(new_ckpt)

        response = self._send_request(request)

        if self._response_validation(response, constants.MSG_SNAP_EXIST):
            LOG.warning("Snapshot %(name)s already exists. "
                        "Skip the creation.",
                        {'name': name})
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to create snapshot %(name)s on "
                         "filesystem %(fs_name)s. Reason: %(err)s.") %
                       {'name': name,
                        'fs_name': fs_name,
                        'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    def get(self, name):
        if name not in self.snap_map:
            request = self._build_query_package(
                self.elt_maker.CheckpointQueryParams(
                    self.elt_maker.Alias(name=name)
                )
            )

            response = self._send_request(request)

            if constants.STATUS_OK != response['maxSeverity']:
                return response['maxSeverity'], response['problems']

            if not response['objects']:
                return constants.STATUS_NOT_FOUND, response['problems']

            src = response['objects'][0]
            snap = {}
            property_map = (
                'name',
                ('id', 'checkpoint'),
                'checkpointOf',
                'state',
            )
            self._copy_properties(src, snap, property_map)

            self.snap_map[name] = snap

        return constants.STATUS_OK, self.snap_map[name]

    def delete(self, name):
        status, out = self.get(name)
        if constants.STATUS_NOT_FOUND == status:
            LOG.warning("Snapshot %s not found. Skip the deletion.",
                        name)
            return
        elif constants.STATUS_OK != status:
            message = (_("Failed to get snapshot by name %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': out})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        chpt_id = self.snap_map[name]['id']

        request = self._build_task_package(
            self.elt_maker.DeleteCheckpoint(checkpoint=chpt_id)
        )

        response = self._send_request(request)
        if constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to delete snapshot %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        self.snap_map.pop(name)

    def get_id(self, name):
        status, out = self.get(name)

        if constants.STATUS_OK != status:
            message = (_("Failed to get snapshot by %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': out})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        return self.snap_map[name]['id']


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class MoverInterface(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(MoverInterface, self).__init__(conn, elt_maker, xml_parser,
                                             manager)

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def create(self, interface):
        # Maximum of 32 characters for mover interface name
        name = interface['name']
        if len(name) > 32:
            name = name[0:31]

        device_name = interface['device_name']
        ip_addr = interface['ip']
        mover_name = interface['mover_name']
        net_mask = interface['net_mask']
        vlan_id = interface['vlan_id'] if interface['vlan_id'] else -1

        mover_id = self._get_mover_id(mover_name, False)

        params = dict(device=device_name,
                      ipAddress=six.text_type(ip_addr),
                      mover=mover_id,
                      name=name,
                      netMask=net_mask,
                      vlanid=six.text_type(vlan_id))

        if interface.get('ip_version') == 6:
            params['ipVersion'] = 'IPv6'

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_task_package(
            self.elt_maker.NewMoverInterface(**params))

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif self._response_validation(
                response, constants.MSG_INTERFACE_NAME_EXIST):
            LOG.warning("Mover interface name %s already exists. "
                        "Skip the creation.", name)
        elif self._response_validation(
                response, constants.MSG_INTERFACE_EXIST):
            LOG.warning("Mover interface IP %s already exists. "
                        "Skip the creation.", ip_addr)
        elif self._response_validation(
                response, constants.MSG_INTERFACE_INVALID_VLAN_ID):
            # When fail to create a mover interface with the specified
            # vlan id, VMAX will leave an interface with vlan id 0 in the
            # backend. So we should explicitly remove the interface.
            try:
                self.delete(six.text_type(ip_addr), mover_name)
            except exception.EMCVmaxXMLAPIError:
                pass
            message = (_("Invalid vlan id %s. Other interfaces on this "
                         "subnet are in a different vlan.") % vlan_id)
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to create mover interface %(interface)s. "
                         "Reason: %(err)s.") %
                       {'interface': interface,
                        'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    def get(self, name, mover_name):
        # Maximum of 32 characters for mover interface name
        if len(name) > 32:
            name = name[0:31]

        status, mover = self.manager.getStorageContext('Mover').get(
            mover_name, True)
        if constants.STATUS_OK == status:
            for interface in mover['interfaces']:
                if name == interface['name']:
                    return constants.STATUS_OK, interface

        return constants.STATUS_NOT_FOUND, None

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def delete(self, ip_addr, mover_name):
        mover_id = self._get_mover_id(mover_name, False)

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_task_package(
            self.elt_maker.DeleteMoverInterface(
                ipAddress=six.text_type(ip_addr),
                mover=mover_id
            )
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif self._response_validation(
                response, constants.MSG_INTERFACE_NON_EXISTENT):
            LOG.warning("Mover interface %s not found. "
                        "Skip the deletion.", ip_addr)
            return
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to delete mover interface %(ip)s on mover "
                         "%(mover)s. Reason: %(err)s.") %
                       {'ip': ip_addr,
                        'mover': mover_name,
                        'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class DNSDomain(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(DNSDomain, self).__init__(conn, elt_maker, xml_parser, manager)

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def create(self, mover_name, name, servers, protocol='udp'):
        mover_id = self._get_mover_id(mover_name, False)

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_task_package(
            self.elt_maker.NewMoverDnsDomain(
                mover=mover_id,
                name=name,
                servers=servers,
                protocol=protocol
            )
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to create DNS domain %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def delete(self, mover_name, name):
        mover_id = self._get_mover_id(mover_name, False)

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_task_package(
            self.elt_maker.DeleteMoverDnsDomain(
                mover=mover_id,
                name=name
            )
        )

        response = self._send_request(request)
        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif constants.STATUS_OK != response['maxSeverity']:
            LOG.warning("Failed to delete DNS domain %(name)s. "
                        "Reason: %(err)s.",
                        {'name': name, 'err': response['problems']})


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class CIFSServer(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(CIFSServer, self).__init__(conn, elt_maker, xml_parser, manager)
        self.cifs_server_map = {}

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def create(self, server_args):
        compName = server_args['name']
        # Maximum of 14 characters for netBIOS name
        name = server_args['name'][-14:]
        # Maximum of 12 characters for alias name
        alias_name = server_args['name'][-12:]
        interfaces = server_args['interface_ip']
        domain_name = server_args['domain_name']
        user_name = server_args['user_name']
        password = server_args['password']
        mover_name = server_args['mover_name']
        is_vdm = server_args['is_vdm']

        mover_id = self._get_mover_id(mover_name, is_vdm)

        if self.xml_retry:
            self.xml_retry = False

        alias_name_list = [self.elt_maker.li(alias_name)]

        request = self._build_task_package(
            self.elt_maker.NewW2KCifsServer(
                self.elt_maker.MoverOrVdm(
                    mover=mover_id,
                    moverIdIsVdm='true' if server_args['is_vdm'] else 'false'
                ),
                self.elt_maker.Aliases(*alias_name_list),
                self.elt_maker.JoinDomain(userName=user_name,
                                          password=password),
                compName=compName,
                domain=domain_name,
                interfaces=interfaces,
                name=name
            )
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        if constants.STATUS_OK != response['maxSeverity']:
            status, out = self.get(compName, mover_name, is_vdm)
            if constants.STATUS_OK == status and out['domainJoined'] == 'true':
                return
            else:
                message = (_("Failed to create CIFS server %(name)s. "
                             "Reason: %(err)s.") %
                           {'name': name,
                            'err': response['problems']})
                LOG.error(message)
                raise exception.EMCVmaxXMLAPIError(err=message)

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def get_all(self, mover_name, is_vdm=True):
        mover_id = self._get_mover_id(mover_name, is_vdm)

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_query_package(
            self.elt_maker.CifsServerQueryParams(
                self.elt_maker.MoverOrVdm(
                    mover=mover_id,
                    moverIdIsVdm='true' if is_vdm else 'false'
                )
            )
        )

        response = self._send_request(request)
        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif constants.STATUS_OK != response['maxSeverity']:
            return response['maxSeverity'], response['objects']

        if mover_name in self.cifs_server_map:
            self.cifs_server_map.pop(mover_name)

        self.cifs_server_map[mover_name] = {}

        for item in response['objects']:
            self.cifs_server_map[mover_name][item['compName'].lower()] = item

        return constants.STATUS_OK, self.cifs_server_map[mover_name]

    def get(self, name, mover_name, is_vdm=True, force=False):
        # name is compName
        name = name.lower()

        if (mover_name in self.cifs_server_map and
                name in self.cifs_server_map[mover_name]) and not force:
            return constants.STATUS_OK, self.cifs_server_map[mover_name][name]

        self.get_all(mover_name, is_vdm)

        if mover_name in self.cifs_server_map:
            for compName, server in self.cifs_server_map[mover_name].items():
                if name == compName:
                    return constants.STATUS_OK, server

        return constants.STATUS_NOT_FOUND, None

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def modify(self, server_args):
        """Make CIFS server join or un-join the domain.

        :param server_args: Dictionary for CIFS server modification
            name: CIFS server name instead of compName
            join_domain: True for joining the domain, false for un-joining
            user_name: User name under which the domain is joined
            password: Password associated with the user name
            mover_name: mover or VDM name
            is_vdm: Boolean to indicate mover or VDM
        :raises exception.EMCVmaxXMLAPIError: if modification fails.
        """
        name = server_args['name']
        join_domain = server_args['join_domain']
        user_name = server_args['user_name']
        password = server_args['password']
        mover_name = server_args['mover_name']

        if 'is_vdm' in server_args.keys():
            is_vdm = server_args['is_vdm']
        else:
            is_vdm = True

        mover_id = self._get_mover_id(mover_name, is_vdm)

        if self.xml_retry:
            self.xml_retry = False

        request = self._build_task_package(
            self.elt_maker.ModifyW2KCifsServer(
                self.elt_maker.DomainSetting(
                    joinDomain='true' if join_domain else 'false',
                    password=password,
                    userName=user_name,
                ),
                mover=mover_id,
                moverIdIsVdm='true' if is_vdm else 'false',
                name=name
            )
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif self._ignore_modification_error(response, join_domain):
            return
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to modify CIFS server %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name,
                        'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    def _ignore_modification_error(self, response, join_domain):
        if self._response_validation(response, constants.MSG_JOIN_DOMAIN):
            return join_domain
        elif self._response_validation(response, constants.MSG_UNJOIN_DOMAIN):
            return not join_domain

        return False

    def delete(self, computer_name, mover_name, is_vdm=True):
        try:
            status, out = self.get(
                computer_name.lower(), mover_name, is_vdm, self.xml_retry)
            if constants.STATUS_NOT_FOUND == status:
                LOG.warning("CIFS server %(name)s on mover %(mover_name)s "
                            "not found. Skip the deletion.",
                            {'name': computer_name, 'mover_name': mover_name})
                return
        except exception.EMCVmaxXMLAPIError:
            LOG.warning("CIFS server %(name)s on mover %(mover_name)s "
                        "not found. Skip the deletion.",
                        {'name': computer_name, 'mover_name': mover_name})
            return

        server_name = out['name']

        mover_id = self._get_mover_id(mover_name, is_vdm)

        request = self._build_task_package(
            self.elt_maker.DeleteCifsServer(
                mover=mover_id,
                moverIdIsVdm='true' if is_vdm else 'false',
                name=server_name
            )
        )

        response = self._send_request(request)

        if constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to delete CIFS server %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': computer_name, 'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        self.cifs_server_map[mover_name].pop(computer_name)


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class CIFSShare(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(CIFSShare, self).__init__(conn, elt_maker, xml_parser, manager)
        self.cifs_share_map = {}

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def create(self, name, server_name, mover_name, is_vdm=True):
        mover_id = self._get_mover_id(mover_name, is_vdm)

        if self.xml_retry:
            self.xml_retry = False

        share_path = '/' + name

        request = self._build_task_package(
            self.elt_maker.NewCifsShare(
                self.elt_maker.MoverOrVdm(
                    mover=mover_id,
                    moverIdIsVdm='true' if is_vdm else 'false'
                ),
                self.elt_maker.CifsServers(self.elt_maker.li(server_name)),
                name=name,
                path=share_path
            )
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to create file share %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    def get(self, name):
        if name not in self.cifs_share_map:
            request = self._build_query_package(
                self.elt_maker.CifsShareQueryParams(name=name)
            )

            response = self._send_request(request)

            if constants.STATUS_OK != response['maxSeverity']:
                return response['maxSeverity'], response['problems']

            if not response['objects']:
                return constants.STATUS_NOT_FOUND, None

            self.cifs_share_map[name] = response['objects'][0]

        return constants.STATUS_OK, self.cifs_share_map[name]

    @utils.retry(exception.EMCVmaxInvalidMoverID)
    def delete(self, name, mover_name, is_vdm=True):
        status, out = self.get(name)
        if constants.STATUS_NOT_FOUND == status:
            LOG.warning("CIFS share %s not found. Skip the deletion.",
                        name)
            return
        elif constants.STATUS_OK != status:
            message = (_("Failed to get CIFS share by name %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': out})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        mover_id = self._get_mover_id(mover_name, is_vdm)

        if self.xml_retry:
            self.xml_retry = False

        netbios_names = self.cifs_share_map[name]['CifsServers']

        request = self._build_task_package(
            self.elt_maker.DeleteCifsShare(
                self.elt_maker.CifsServers(*map(lambda a: self.elt_maker.li(a),
                                                netbios_names)),
                mover=mover_id,
                moverIdIsVdm='true' if is_vdm else 'false',
                name=name
            )
        )

        response = self._send_request(request)

        if (self._response_validation(response,
                                      constants.MSG_INVALID_MOVER_ID) and
                not self.xml_retry):
            self.xml_retry = True
            raise exception.EMCVmaxInvalidMoverID(id=mover_id)
        elif constants.STATUS_OK != response['maxSeverity']:
            message = (_("Failed to delete file system %(name)s. "
                         "Reason: %(err)s.") %
                       {'name': name, 'err': response['problems']})
            LOG.error(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        self.cifs_share_map.pop(name)

    def disable_share_access(self, share_name, mover_name):
        cmd_str = 'sharesd %s set noaccess' % share_name
        disable_access = [
            'env', 'NAS_DB=/nas', '/nas/bin/.server_config', mover_name,
            '-v', "%s" % cmd_str,
        ]

        try:
            self._execute_cmd(disable_access, check_exit_code=True)
        except processutils.ProcessExecutionError:
            message = (_('Failed to disable the access to CIFS share '
                         '%(name)s.') %
                       {'name': share_name})
            LOG.exception(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    def allow_share_access(self, mover_name, share_name, user_name, domain,
                           access=constants.CIFS_ACL_FULLCONTROL):
        account = user_name + "@" + domain
        allow_str = ('sharesd %(share_name)s grant %(account)s=%(access)s'
                     % {'share_name': share_name,
                        'account': account,
                        'access': access})

        allow_access = [
            'env', 'NAS_DB=/nas', '/nas/bin/.server_config', mover_name,
            '-v', "%s" % allow_str,
        ]

        try:
            self._execute_cmd(allow_access, check_exit_code=True)
        except processutils.ProcessExecutionError as expt:
            dup_msg = re.compile(r'ACE for %(domain)s\\%(user)s unchanged' %
                                 {'domain': domain, 'user': user_name}, re.I)
            if re.search(dup_msg, expt.stdout):
                LOG.warning("Duplicate access control entry, "
                            "skipping allow...")
            else:
                message = (_('Failed to allow the access %(access)s to '
                             'CIFS share %(name)s. Reason: %(err)s.') %
                           {'access': access, 'name': share_name, 'err': expt})
                LOG.error(message)
                raise exception.EMCVmaxXMLAPIError(err=message)

    def deny_share_access(self, mover_name, share_name, user_name, domain,
                          access=constants.CIFS_ACL_FULLCONTROL):
        account = user_name + "@" + domain
        revoke_str = ('sharesd %(share_name)s revoke %(account)s=%(access)s'
                      % {'share_name': share_name,
                         'account': account,
                         'access': access})

        allow_access = [
            'env', 'NAS_DB=/nas', '/nas/bin/.server_config', mover_name,
            '-v', "%s" % revoke_str,
        ]
        try:
            self._execute_cmd(allow_access, check_exit_code=True)
        except processutils.ProcessExecutionError as expt:
            not_found_msg = re.compile(
                r'No ACE found for %(domain)s\\%(user)s'
                % {'domain': domain, 'user': user_name}, re.I)
            user_err_msg = re.compile(
                r'Cannot get mapping for %(domain)s\\%(user)s'
                % {'domain': domain, 'user': user_name}, re.I)

            if re.search(not_found_msg, expt.stdout):
                LOG.warning("No access control entry found, "
                            "skipping deny...")
            elif re.search(user_err_msg, expt.stdout):
                LOG.warning("User not found on domain, skipping deny...")
            else:
                message = (_('Failed to deny the access %(access)s to '
                             'CIFS share %(name)s. Reason: %(err)s.') %
                           {'access': access, 'name': share_name, 'err': expt})
                LOG.exception(message)
                raise exception.EMCVmaxXMLAPIError(err=message)

    def get_share_access(self, mover_name, share_name):
        get_str = 'sharesd %s dump' % share_name
        get_access = [
            'env', 'NAS_DB=/nas', '/nas/bin/.server_config', mover_name,
            '-v', "%s" % get_str,
        ]

        try:
            out, err = self._execute_cmd(get_access, check_exit_code=True)
        except processutils.ProcessExecutionError:
            msg = _('Failed to get access list of CIFS share %s.') % share_name
            LOG.exception(msg)
            raise exception.EMCVmaxXMLAPIError(err=msg)

        ret = {}
        name_pattern = re.compile(r"Unix user '(.+?)'")
        access_pattern = re.compile(r"ALLOWED:(.+?):")

        name = None
        for line in out.splitlines():
            if name is None:
                names = name_pattern.findall(line)
                if names:
                    name = names[0].lower()
            else:
                accesses = access_pattern.findall(line)
                if accesses:
                    ret[name] = accesses[0].lower()
                    name = None
        return ret

    def clear_share_access(self, mover_name, share_name, domain,
                           white_list_users):
        existing_users = self.get_share_access(mover_name, share_name)
        white_list_users_set = set(user.lower() for user in white_list_users)
        users_to_remove = set(existing_users.keys()) - white_list_users_set
        for user in users_to_remove:
            self.deny_share_access(mover_name, share_name, user, domain,
                                   existing_users[user])
        return users_to_remove


@vmax_utils.decorate_all_methods(vmax_utils.log_enter_exit,
                                 debug_only=True)
class NFSShare(StorageObject):
    def __init__(self, conn, elt_maker, xml_parser, manager):
        super(NFSShare, self).__init__(conn, elt_maker, xml_parser, manager)
        self.nfs_share_map = {}

    def create(self, name, mover_name):
        share_path = '/' + name
        create_nfs_share_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/server_export', mover_name,
            '-option', 'access=-0.0.0.0/0.0.0.0',
            share_path,
        ]

        try:
            self._execute_cmd(create_nfs_share_cmd, check_exit_code=True)
        except processutils.ProcessExecutionError as expt:
            message = (_('Failed to create NFS share %(name)s on mover '
                         '%(mover_name)s. Reason: %(err)s.') %
                       {'name': name, 'mover_name': mover_name, 'err': expt})
            LOG.exception(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

    def delete(self, name, mover_name):
        path = '/' + name

        status, out = self.get(name, mover_name)
        if constants.STATUS_NOT_FOUND == status:
            LOG.warning("NFS share %s not found. Skip the deletion.",
                        path)
            return

        delete_nfs_share_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/server_export', mover_name,
            '-unexport',
            '-perm',
            path,
        ]

        try:
            self._execute_cmd(delete_nfs_share_cmd, check_exit_code=True)
        except processutils.ProcessExecutionError as expt:
            message = (_('Failed to delete NFS share %(name)s on '
                         '%(mover_name)s. Reason: %(err)s.') %
                       {'name': name, 'mover_name': mover_name, 'err': expt})
            LOG.exception(message)
            raise exception.EMCVmaxXMLAPIError(err=message)

        self.nfs_share_map.pop(name)

    def get(self, name, mover_name, force=False, check_exit_code=False):
        if name in self.nfs_share_map and not force:
            return constants.STATUS_OK, self.nfs_share_map[name]

        path = '/' + name

        nfs_share = {
            "mover_name": '',
            "path": '',
            'AccessHosts': [],
            'RwHosts': [],
            'RoHosts': [],
            'RootHosts': [],
            'readOnly': '',
        }

        nfs_query_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/server_export', mover_name,
            '-P', 'nfs',
            '-list', path,
        ]

        try:
            out, err = self._execute_cmd(nfs_query_cmd,
                                         check_exit_code=check_exit_code)
        except processutils.ProcessExecutionError as expt:
            dup_msg = (r'%(mover_name)s : No such file or directory' %
                       {'mover_name': mover_name})
            if re.search(dup_msg, expt.stdout):
                LOG.warning("NFS share %s not found.", name)
                return constants.STATUS_NOT_FOUND, None
            else:
                message = (_('Failed to list NFS share %(name)s on '
                             '%(mover_name)s. Reason: %(err)s.') %
                           {'name': name,
                            'mover_name': mover_name,
                            'err': expt})
                LOG.exception(message)
                raise exception.EMCVmaxXMLAPIError(err=message)

        re_exports = '%s\s*:\s*\nexport\s*(.*)\n' % mover_name
        m = re.search(re_exports, out)
        if m is not None:
            nfs_share['path'] = path
            nfs_share['mover_name'] = mover_name
            export = m.group(1)
            fields = export.split(" ")
            for field in fields:
                field = field.strip()
                if field.startswith('rw='):
                    nfs_share['RwHosts'] = vmax_utils.parse_ipaddr(field[3:])
                elif field.startswith('access='):
                    nfs_share['AccessHosts'] = vmax_utils.parse_ipaddr(
                        field[7:])
                elif field.startswith('root='):
                    nfs_share['RootHosts'] = vmax_utils.parse_ipaddr(field[5:])
                elif field.startswith('ro='):
                    nfs_share['RoHosts'] = vmax_utils.parse_ipaddr(field[3:])

            self.nfs_share_map[name] = nfs_share
        else:
            return constants.STATUS_NOT_FOUND, None

        return constants.STATUS_OK, self.nfs_share_map[name]

    def allow_share_access(self, share_name, host_ip, mover_name,
                           access_level=const.ACCESS_LEVEL_RW):
        @utils.synchronized('emc-shareaccess-' + share_name)
        def do_allow_access(share_name, host_ip, mover_name, access_level):
            status, share = self.get(share_name, mover_name)
            if constants.STATUS_NOT_FOUND == status:
                message = (_('NFS share %s not found.') % share_name)
                LOG.error(message)
                raise exception.EMCVmaxXMLAPIError(err=message)

            changed = False
            rwhosts = share['RwHosts']
            rohosts = share['RoHosts']

            host_ip = vmax_utils.convert_ipv6_format_if_needed(host_ip)

            if access_level == const.ACCESS_LEVEL_RW:
                if host_ip not in rwhosts:
                    rwhosts.append(host_ip)
                    changed = True
                if host_ip in rohosts:
                    rohosts.remove(host_ip)
                    changed = True
            if access_level == const.ACCESS_LEVEL_RO:
                if host_ip not in rohosts:
                    rohosts.append(host_ip)
                    changed = True
                if host_ip in rwhosts:
                    rwhosts.remove(host_ip)
                    changed = True

            roothosts = share['RootHosts']
            if host_ip not in roothosts:
                roothosts.append(host_ip)
                changed = True
            accesshosts = share['AccessHosts']
            if host_ip not in accesshosts:
                accesshosts.append(host_ip)
                changed = True

            if not changed:
                LOG.debug("%(host)s is already in access list of share "
                          "%(name)s.", {'host': host_ip, 'name': share_name})
            else:
                path = '/' + share_name
                self._set_share_access(path,
                                       mover_name,
                                       rwhosts,
                                       rohosts,
                                       roothosts,
                                       accesshosts)

                # Update self.nfs_share_map
                self.get(share_name, mover_name, force=True,
                         check_exit_code=True)

        do_allow_access(share_name, host_ip, mover_name, access_level)

    def deny_share_access(self, share_name, host_ip, mover_name):

        @utils.synchronized('emc-shareaccess-' + share_name)
        def do_deny_access(share_name, host_ip, mover_name):
            status, share = self.get(share_name, mover_name)
            if constants.STATUS_OK != status:
                message = (_('Query nfs share %(path)s failed. '
                             'Reason %(err)s.') %
                           {'path': share_name, 'err': share})
                LOG.error(message)
                raise exception.EMCVmaxXMLAPIError(err=message)

            changed = False
            rwhosts = set(share['RwHosts'])
            if host_ip in rwhosts:
                rwhosts.remove(host_ip)
                changed = True
            roothosts = set(share['RootHosts'])
            if host_ip in roothosts:
                roothosts.remove(host_ip)
                changed = True
            accesshosts = set(share['AccessHosts'])
            if host_ip in accesshosts:
                accesshosts.remove(host_ip)
                changed = True
            rohosts = set(share['RoHosts'])
            if host_ip in rohosts:
                rohosts.remove(host_ip)
                changed = True
            if not changed:
                LOG.debug("%(host)s is already in access list of share "
                          "%(name)s.", {'host': host_ip, 'name': share_name})
            else:
                path = '/' + share_name
                self._set_share_access(path,
                                       mover_name,
                                       rwhosts,
                                       rohosts,
                                       roothosts,
                                       accesshosts)

                # Update self.nfs_share_map
                self.get(share_name, mover_name, force=True,
                         check_exit_code=True)

        do_deny_access(share_name, host_ip, mover_name)

    def clear_share_access(self, share_name, mover_name, white_list_hosts):
        @utils.synchronized('emc-shareaccess-' + share_name)
        def do_clear_access(share_name, mover_name, white_list_hosts):
            def hosts_to_remove(orig_list):
                if white_list_hosts is None:
                    ret = set()
                else:
                    ret = set(white_list_hosts).intersection(set(orig_list))
                return ret

            status, share = self.get(share_name, mover_name)
            if constants.STATUS_OK != status:
                message = (_('Query nfs share %(path)s failed. '
                             'Reason %(err)s.') %
                           {'path': share_name, 'err': status})
                raise exception.EMCVmaxXMLAPIError(err=message)

            self._set_share_access('/' + share_name,
                                   mover_name,
                                   hosts_to_remove(share['RwHosts']),
                                   hosts_to_remove(share['RoHosts']),
                                   hosts_to_remove(share['RootHosts']),
                                   hosts_to_remove(share['AccessHosts']))

            # Update self.nfs_share_map
            self.get(share_name, mover_name, force=True,
                     check_exit_code=True)

        do_clear_access(share_name, mover_name, white_list_hosts)

    def _set_share_access(self, path, mover_name, rw_hosts, ro_hosts,
                          root_hosts, access_hosts):

        if access_hosts is None:
            access_hosts = set()

        if '-0.0.0.0/0.0.0.0' not in access_hosts:
            access_hosts.add('-0.0.0.0/0.0.0.0')

        access_str = ('access=%(access)s'
                      % {'access': ':'.join(access_hosts)})
        if root_hosts:
            access_str += (',root=%(root)s' % {'root': ':'.join(root_hosts)})
        if rw_hosts:
            access_str += ',rw=%(rw)s' % {'rw': ':'.join(rw_hosts)}
        if ro_hosts:
            access_str += ',ro=%(ro)s' % {'ro': ':'.join(ro_hosts)}
        set_nfs_share_access_cmd = [
            'env', 'NAS_DB=/nas', '/nas/bin/server_export', mover_name,
            '-ignore',
            '-option', access_str,
            path,
        ]

        try:
            self._execute_cmd(set_nfs_share_access_cmd, check_exit_code=True)
        except processutils.ProcessExecutionError as expt:
            message = (_('Failed to set NFS share %(name)s access on '
                         '%(mover_name)s. Reason: %(err)s.') %
                       {'name': path[1:],
                        'mover_name': mover_name,
                        'err': expt})
            LOG.exception(message)
            raise exception.EMCVmaxXMLAPIError(err=message)



# Copyright 2012 NetApp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the LVM driver module."""

import os

import ddt
import mock
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_utils import timeutils

from manila.common import constants as const
from manila import context
from manila import exception
from manila.share import configuration
from manila.share.drivers import lvm
from manila import test
from manila.tests.db import fakes as db_fakes
from manila.tests import fake_utils
from manila.tests.share.drivers import test_generic


CONF = cfg.CONF


def fake_share(**kwargs):
    share = {
        'id': 'fakeid',
        'name': 'fakename',
        'size': 1,
        'share_proto': 'NFS',
        'export_location': '127.0.0.1:/mnt/nfs/volume-00002',
    }
    share.update(kwargs)
    return db_fakes.FakeModel(share)


def fake_snapshot(**kwargs):
    snapshot = {
        'id': 'fakesnapshotid',
        'share_name': 'fakename',
        'share_id': 'fakeid',
        'name': 'fakesnapshotname',
        'share_proto': 'NFS',
        'export_location': '127.0.0.1:/mnt/nfs/volume-00002',
        'share': {
            'id': 'fakeid',
            'name': 'fakename',
            'size': 1,
            'share_proto': 'NFS',
        },
    }
    snapshot.update(kwargs)
    return db_fakes.FakeModel(snapshot)


def fake_access(**kwargs):
    access = {
        'id': 'fakeaccid',
        'access_type': 'ip',
        'access_to': '10.0.0.2',
        'access_level': 'rw',
        'state': 'active',
    }
    access.update(kwargs)
    return db_fakes.FakeModel(access)


@ddt.ddt
class LVMShareDriverTestCase(test.TestCase):
    """Tests LVMShareDriver."""

    def setUp(self):
        super(LVMShareDriverTestCase, self).setUp()
        fake_utils.stub_out_utils_execute(self)
        self._context = context.get_admin_context()

        CONF.set_default('lvm_share_volume_group', 'fakevg')
        CONF.set_default('lvm_share_export_ips', ['10.0.0.1', '10.0.0.2'])
        CONF.set_default('driver_handles_share_servers', False)
        CONF.set_default('reserved_share_percentage', 50)

        self._helper_cifs = mock.Mock()
        self._helper_nfs = mock.Mock()
        self.fake_conf = configuration.Configuration(None)
        self._db = mock.Mock()
        self._os = lvm.os = mock.Mock()
        self._os.path.join = os.path.join
        self._driver = lvm.LVMShareDriver(self._db,
                                          configuration=self.fake_conf)
        self._driver._helpers = {
            'CIFS': self._helper_cifs,
            'NFS': self._helper_nfs,
        }

        self.share = fake_share()
        self.access = fake_access()
        self.snapshot = fake_snapshot()
        self.server = {
            'public_addresses': self.fake_conf.lvm_share_export_ips,
            'instance_id': 'LVM',
            'lock_name': 'manila_lvm',
        }

        # Used only to test compatibility with share manager
        self.share_server = "fake_share_server"

    def tearDown(self):
        super(LVMShareDriverTestCase, self).tearDown()
        fake_utils.fake_execute_set_repliers([])
        fake_utils.fake_execute_clear_log()

    def test_do_setup(self):
        CONF.set_default('lvm_share_helpers', ['NFS=fakenfs'])
        lvm.importutils = mock.Mock()
        lvm.importutils.import_class.return_value = self._helper_nfs
        self._driver.do_setup(self._context)
        lvm.importutils.import_class.assert_has_calls([
            mock.call('fakenfs')
        ])

    def test_check_for_setup_error(self):
        def exec_runner(*ignore_args, **ignore_kwargs):
            return '\n   fake1\n   fakevg\n   fake2\n', ''

        expected_exec = ['vgs --noheadings -o name']
        fake_utils.fake_execute_set_repliers([(expected_exec[0], exec_runner)])
        self._driver.check_for_setup_error()
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())

    def test_check_for_setup_error_no_vg(self):
        def exec_runner(*ignore_args, **ignore_kwargs):
            return '\n   fake0\n   fake1\n   fake2\n', ''

        fake_utils.fake_execute_set_repliers([('vgs --noheadings -o name',
                                               exec_runner)])
        self.assertRaises(exception.InvalidParameterValue,
                          self._driver.check_for_setup_error)

    def test_check_for_setup_error_deprecated_export_ip(self):
        def exec_runner(*ignore_args, **ignore_kwargs):
            return '\n   fake1\n   fakevg\n   fake2\n', ''

        fake_utils.fake_execute_set_repliers([('vgs --noheadings -o name',
                                               exec_runner)])
        CONF.set_default('lvm_share_export_ip', CONF.lvm_share_export_ips[0])
        CONF.set_default('lvm_share_export_ips', None)
        self.assertIsNone(self._driver.check_for_setup_error())

    def test_check_for_setup_error_no_export_ips(self):
        def exec_runner(*ignore_args, **ignore_kwargs):
            return '\n   fake1\n   fakevg\n   fake2\n', ''

        fake_utils.fake_execute_set_repliers([('vgs --noheadings -o name',
                                               exec_runner)])
        CONF.set_default('lvm_share_export_ips', None)
        self.assertRaises(exception.InvalidParameterValue,
                          self._driver.check_for_setup_error)

    def test_check_for_setup_error_both_export_ip_and_ips(self):
        def exec_runner(*ignore_args, **ignore_kwargs):
            return '\n   fake1\n   fakevg\n   fake2\n', ''

        fake_utils.fake_execute_set_repliers([('vgs --noheadings -o name',
                                               exec_runner)])
        CONF.set_default('lvm_share_export_ip', CONF.lvm_share_export_ips[0])
        self.assertRaises(exception.InvalidParameterValue,
                          self._driver.check_for_setup_error)

    def test_local_path_normal(self):
        share = fake_share(name='fake_sharename')
        CONF.set_default('lvm_share_volume_group', 'fake_vg')
        ret = self._driver._get_local_path(share)
        self.assertEqual('/dev/mapper/fake_vg-fake_sharename', ret)

    def test_local_path_escapes(self):
        share = fake_share(name='fake-sharename')
        CONF.set_default('lvm_share_volume_group', 'fake-vg')
        ret = self._driver._get_local_path(share)
        self.assertEqual('/dev/mapper/fake--vg-fake--sharename', ret)

    def test_create_share(self):
        CONF.set_default('lvm_share_mirrors', 0)
        self._driver._mount_device = mock.Mock()

        ret = self._driver.create_share(self._context, self.share,
                                        self.share_server)

        self._driver._mount_device.assert_called_with(
            self.share, '/dev/mapper/fakevg-fakename')
        expected_exec = [
            'lvcreate -L 1G -n fakename fakevg',
            'mkfs.ext4 /dev/mapper/fakevg-fakename',
        ]
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())
        self.assertEqual(self._helper_nfs.create_exports.return_value, ret)

    def test_create_share_from_snapshot(self):
        CONF.set_default('lvm_share_mirrors', 0)
        self._driver._mount_device = mock.Mock()
        snapshot_instance = {
            'snapshot_id': 'fakesnapshotid',
            'name': 'fakename'
        }
        mount_share = '/dev/mapper/fakevg-fakename'
        mount_snapshot = '/dev/mapper/fakevg-fakename'
        self._helper_nfs.create_export.return_value = 'fakelocation'
        self._driver.create_share_from_snapshot(self._context,
                                                self.share,
                                                snapshot_instance,
                                                self.share_server)

        self._driver._mount_device.assert_called_with(self.share,
                                                      mount_snapshot)
        expected_exec = [
            'lvcreate -L 1G -n fakename fakevg',
            'mkfs.ext4 /dev/mapper/fakevg-fakename',
            'e2fsck -y -f %s' % mount_share,
            'tune2fs -U random %s' % mount_share,
            ("dd count=0 if=%s of=%s iflag=direct oflag=direct" %
             (mount_snapshot, mount_share)),
            ("dd if=%s of=%s count=1024 bs=1M iflag=direct oflag=direct" %
             (mount_snapshot, mount_share)),
        ]
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())

    def test_create_share_mirrors(self):
        share = fake_share(size='2048')
        CONF.set_default('lvm_share_mirrors', 2)
        self._driver._mount_device = mock.Mock()

        ret = self._driver.create_share(self._context, share,
                                        self.share_server)

        self._driver._mount_device.assert_called_with(
            share, '/dev/mapper/fakevg-fakename')
        expected_exec = [
            'lvcreate -L 2048G -n fakename fakevg -m 2 --nosync -R 2',
            'mkfs.ext4 /dev/mapper/fakevg-fakename',
        ]
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())
        self.assertEqual(self._helper_nfs.create_exports.return_value, ret)

    def test_deallocate_container(self):
        expected_exec = ['lvremove -f fakevg/fakename']
        self._driver._deallocate_container(self.share['name'])
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())

    def test_deallocate_container_error(self):
        def _fake_exec(*args, **kwargs):
            raise exception.ProcessExecutionError(stderr="error")

        self.mock_object(self._driver, '_try_execute', _fake_exec)
        self.assertRaises(exception.ProcessExecutionError,
                          self._driver._deallocate_container,
                          self.share['name'])

    def test_deallocate_container_not_found_error(self):
        def _fake_exec(*args, **kwargs):
            raise exception.ProcessExecutionError(stderr="not found")

        self.mock_object(self._driver, '_try_execute', _fake_exec)
        self._driver._deallocate_container(self.share['name'])

    @mock.patch.object(lvm.LVMShareDriver, '_update_share_stats', mock.Mock())
    def test_get_share_stats(self):
        with mock.patch.object(self._driver, '_stats', mock.Mock) as stats:
            self.assertEqual(stats, self._driver.get_share_stats())
        self.assertFalse(self._driver._update_share_stats.called)

    @mock.patch.object(lvm.LVMShareDriver, '_update_share_stats', mock.Mock())
    def test_get_share_stats_refresh(self):
        with mock.patch.object(self._driver, '_stats', mock.Mock) as stats:
            self.assertEqual(stats,
                             self._driver.get_share_stats(refresh=True))
        self._driver._update_share_stats.assert_called_once_with()

    def test__unmount_device_is_busy_error(self):
        def exec_runner(*ignore_args, **ignore_kwargs):
            raise exception.ProcessExecutionError(stderr='device is busy')
        self._os.path.exists.return_value = True
        mount_path = self._get_mount_path(self.share)
        expected_exec = [
            "umount -f %s" % (mount_path),
        ]
        fake_utils.fake_execute_set_repliers([(expected_exec[0], exec_runner)])

        self.assertRaises(exception.ShareBusyException,
                          self._driver._unmount_device,
                          self.share)
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())

    def test__unmount_device_error(self):
        def exec_runner(*ignore_args, **ignore_kwargs):
            raise exception.ProcessExecutionError(stderr='fake error')
        mount_path = self._get_mount_path(self.share)
        self._os.path.exists.return_value = True
        cmd = "umount -f %s" % (mount_path)
        fake_utils.fake_execute_set_repliers([(cmd, exec_runner)])
        self.assertRaises(processutils.ProcessExecutionError,
                          self._driver._unmount_device,
                          self.share)
        self._os.path.exists.assert_called_with(mount_path)

    def test__unmount_device_rmdir_error(self):
        def exec_runner(*ignore_args, **ignore_kwargs):
            raise exception.ProcessExecutionError(stderr='fake error')
        mount_path = self._get_mount_path(self.share)
        self._os.path.exists.return_value = True
        cmd = "rmdir %s" % (mount_path)
        fake_utils.fake_execute_set_repliers([(cmd, exec_runner)])
        self.assertRaises(processutils.ProcessExecutionError,
                          self._driver._unmount_device,
                          self.share)
        self._os.path.exists.assert_called_with(mount_path)

    def test_create_snapshot(self):
        self._driver.create_snapshot(self._context, self.snapshot,
                                     self.share_server)
        mount_path = self._get_mount_path(self.snapshot)
        expected_exec = [
            ("lvcreate -L 1G --name fakesnapshotname --snapshot "
             "%s/fakename" % (CONF.lvm_share_volume_group,)),
            "e2fsck -y -f /dev/mapper/fakevg-%s" % self.snapshot['name'],
            "tune2fs -U random /dev/mapper/fakevg-%s" % self.snapshot['name'],
            "mkdir -p " + mount_path,
            "mount /dev/mapper/fakevg-fakesnapshotname " + mount_path,
            "chmod 777 " + mount_path,
        ]
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())

    def test_ensure_share(self):
        device_name = '/dev/mapper/fakevg-fakename'
        with mock.patch.object(self._driver,
                               '_mount_device',
                               mock.Mock(return_value='fake_location')):
            self._driver.ensure_share(self._context, self.share,
                                      self.share_server)
            self._driver._mount_device.assert_called_with(self.share,
                                                          device_name)
            self._helper_nfs.create_exports.assert_called_once_with(
                self.server, self.share['name'], recreate=True)

    def test_delete_share(self):
        mount_path = self._get_mount_path(self.share)
        self._helper_nfs.remove_export(mount_path, self.share['name'])
        self._driver._delete_share(self._context, self.share)

    def test_delete_snapshot(self):
        mount_path = self._get_mount_path(self.snapshot)
        expected_exec = [
            'umount -f %s' % mount_path,
            'rmdir %s' % mount_path,
            'lvremove -f fakevg/fakesnapshotname',
        ]
        self._driver.delete_snapshot(self._context, self.snapshot,
                                     self.share_server)
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())

    def test_delete_share_invalid_share(self):
        self._driver._get_helper = mock.Mock(
            side_effect=exception.InvalidShare(reason='fake'))
        self._driver.delete_share(self._context, self.share, self.share_server)

    def test_delete_share_process_execution_error(self):
        self.mock_object(
            self._helper_nfs,
            'remove_export',
            mock.Mock(side_effect=exception.ProcessExecutionError))

        self._driver._delete_share(self._context, self.share)
        self._helper_nfs.remove_exports.assert_called_once_with(
            self.server,
            self.share['name'])

    @ddt.data(const.ACCESS_LEVEL_RW, const.ACCESS_LEVEL_RO)
    def test_update_access(self, access_level):
        access_rules = [test_generic.get_fake_access_rule(
            '1.1.1.1', access_level), ]
        add_rules = [test_generic.get_fake_access_rule(
            '2.2.2.2', access_level), ]
        delete_rules = [test_generic.get_fake_access_rule(
            '3.3.3.3', access_level), ]
        self._driver.update_access(self._context, self.share, access_rules,
                                   add_rules=add_rules,
                                   delete_rules=delete_rules,
                                   share_server=self.server)
        (self._driver._helpers[self.share['share_proto']].
            update_access.assert_called_once_with(
                self.server, self.share['name'],
                access_rules, add_rules=add_rules, delete_rules=delete_rules))

    @ddt.data(('1001::1001/129', None, False), ('1.1.1.256', None, False),
              ('1001::1001', None, [6]), ('1.1.1.0', None, [4]),
              (None, ['1001::1001', '1.1.1.0'], [6, 4]),
              (None, ['1001::1001'], [6]), (None, ['1.1.1.0'], [4]),
              (None, ['1001::1001/129', '1.1.1.0'], False))
    @ddt.unpack
    def test_get_configured_ip_versions(
            self, configured_ip, configured_ips, configured_ip_version):
        CONF.set_default('lvm_share_export_ip', configured_ip)
        CONF.set_default('lvm_share_export_ips', configured_ips)
        if configured_ip_version:
            self.assertEqual(configured_ip_version,
                             self._driver.get_configured_ip_versions())
        else:
            self.assertRaises(exception.InvalidInput,
                              self._driver.get_configured_ip_versions)

    def test_mount_device(self):
        mount_path = self._get_mount_path(self.share)
        ret = self._driver._mount_device(self.share, 'fakedevice')
        expected_exec = [
            "mkdir -p %s" % (mount_path,),
            "mount fakedevice %s" % (mount_path,),
            "chmod 777 %s" % (mount_path,),
        ]
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())
        self.assertEqual(mount_path, ret)

    def test_mount_device_already(self):
        def exec_runner(*args, **kwargs):
            if 'mount' in args and '-l' not in args:
                raise exception.ProcessExecutionError()
            else:
                return 'fakedevice', ''

        self.mock_object(self._driver, '_execute', exec_runner)
        mount_path = self._get_mount_path(self.share)

        ret = self._driver._mount_device(self.share, 'fakedevice')
        self.assertEqual(mount_path, ret)

    def test_mount_device_error(self):
        def exec_runner(*args, **kwargs):
            if 'mount' in args and '-l' not in args:
                raise exception.ProcessExecutionError()
            else:
                return 'fake', ''

        self.mock_object(self._driver, '_execute', exec_runner)
        self.assertRaises(exception.ProcessExecutionError,
                          self._driver._mount_device, self.share, 'fakedevice')

    def test_get_helper(self):
        share_cifs = fake_share(share_proto='CIFS')
        share_nfs = fake_share(share_proto='NFS')
        share_fake = fake_share(share_proto='FAKE')
        self.assertEqual(self._driver._get_helper(share_cifs),
                         self._helper_cifs)
        self.assertEqual(self._driver._get_helper(share_nfs),
                         self._helper_nfs)
        self.assertRaises(exception.InvalidShare, self._driver._get_helper,
                          share_fake)

    def _get_mount_path(self, share):
        return os.path.join(CONF.lvm_share_export_root, share['name'])

    def test__unmount_device(self):
        mount_path = self._get_mount_path(self.share)
        self._os.path.exists.return_value = True
        self.mock_object(self._driver, '_execute')
        self._driver._unmount_device(self.share)
        self._driver._execute.assert_any_call('umount', '-f', mount_path,
                                              run_as_root=True)
        self._driver._execute.assert_any_call('rmdir', mount_path,
                                              run_as_root=True)
        self._os.path.exists.assert_called_with(mount_path)

    def test_extend_share(self):
        local_path = self._driver._get_local_path(self.share)
        self.mock_object(self._driver, '_extend_container')
        self.mock_object(self._driver, '_execute')
        self._driver.extend_share(self.share, 3)
        self._driver._extend_container.assert_called_once_with(self.share,
                                                               local_path, 3)
        self._driver._execute.assert_called_once_with('resize2fs', local_path,
                                                      run_as_root=True)

    def test_ssh_exec_as_root(self):
        command = ['fake_command']
        self.mock_object(self._driver, '_execute')
        self._driver._ssh_exec_as_root('fake_server', command)
        self._driver._execute.assert_called_once_with('fake_command',
                                                      check_exit_code=True)

    def test_ssh_exec_as_root_with_sudo(self):
        command = ['sudo', 'fake_command']
        self.mock_object(self._driver, '_execute')
        self._driver._ssh_exec_as_root('fake_server', command)
        self._driver._execute.assert_called_once_with(
            'fake_command', run_as_root=True, check_exit_code=True)

    def test_extend_container(self):
        self.mock_object(self._driver, '_try_execute')
        self._driver._extend_container(self.share, 'device_name', 3)
        self._driver._try_execute.assert_called_once_with(
            'lvextend',
            '-L',
            '3G',
            '-n',
            'device_name',
            run_as_root=True)

    def test_get_share_server_pools(self):
        expected_result = [{
            'pool_name': 'lvm-single-pool',
            'total_capacity_gb': 33,
            'free_capacity_gb': 22,
            'reserved_percentage': 0,
        }, ]
        self.mock_object(
            self._driver,
            '_execute',
            mock.Mock(return_value=("VSize 33g VFree 22g", None)))

        self.assertEqual(expected_result,
                         self._driver.get_share_server_pools())
        self._driver._execute.assert_called_once_with(
            'vgs', 'fakevg', '--rows', '--units', 'g', run_as_root=True)

    def test_copy_volume_error(self):
        def _fake_exec(*args, **kwargs):
            if 'count=0' in args:
                raise exception.ProcessExecutionError()

        self.mock_object(self._driver, '_execute',
                         mock.Mock(side_effect=_fake_exec))
        self._driver._copy_volume('src', 'dest', 1)
        self._driver._execute.assert_any_call('dd', 'count=0', 'if=src',
                                              'of=dest', 'iflag=direct',
                                              'oflag=direct', run_as_root=True)
        self._driver._execute.assert_any_call('dd', 'if=src', 'of=dest',
                                              'count=1024', 'bs=1M',
                                              run_as_root=True)

    @ddt.data(('1.1.1.1', 4), ('1001::1001', 6))
    @ddt.unpack
    def test_update_share_stats(self, configured_ip, version):
        CONF.set_default('lvm_share_export_ip', configured_ip)
        self.mock_object(self._driver, 'get_share_server_pools',
                         mock.Mock(return_value='test-pool'))

        self._driver._update_share_stats()
        self.assertEqual('LVM', self._driver._stats['share_backend_name'])
        self.assertEqual('NFS_CIFS', self._driver._stats['storage_protocol'])
        self.assertEqual(50, self._driver._stats['reserved_percentage'])
        self.assertTrue(self._driver._stats['snapshot_support'])
        self.assertEqual('LVMShareDriver', self._driver._stats['driver_name'])
        self.assertEqual('test-pool', self._driver._stats['pools'])
        self.assertEqual(version == 4, self._driver._stats['ipv4_support'])
        self.assertEqual(version == 6, self._driver._stats['ipv6_support'])

    def test_revert_to_snapshot(self):
        mock_update_access = self.mock_object(self._helper_nfs,
                                              'update_access')
        self._driver.revert_to_snapshot(self._context, self.snapshot,
                                        [], [], self.share_server)
        snap_lv = "%s/fakesnapshotname" % (CONF.lvm_share_volume_group)
        share_lv = "%s/fakename" % (CONF.lvm_share_volume_group)
        share_mount_path = self._get_mount_path(self.snapshot['share'])
        snapshot_mount_path = self._get_mount_path(self.snapshot)
        expected_exec = [
            ('umount -f %s' % snapshot_mount_path),
            ("rmdir %s" % snapshot_mount_path),
            ("umount -f %s" % share_mount_path),
            ("rmdir %s" % share_mount_path),
            ("lvconvert --merge %s" % snap_lv),
            ("lvcreate -L 1G --name fakesnapshotname --snapshot %s" %
                share_lv),
            ("e2fsck -y -f /dev/mapper/%s-fakesnapshotname" %
                CONF.lvm_share_volume_group),
            ("tune2fs -U random /dev/mapper/%s-fakesnapshotname" %
                CONF.lvm_share_volume_group),
            ("mkdir -p %s" % share_mount_path),
            ("mount /dev/mapper/%s-fakename %s" %
                (CONF.lvm_share_volume_group, share_mount_path)),
            ("chmod 777 %s" % share_mount_path),
            ("mkdir -p %s" % snapshot_mount_path),
            ("mount /dev/mapper/fakevg-fakesnapshotname "
             "%s" % snapshot_mount_path),
            ("chmod 777 %s" % snapshot_mount_path),
        ]
        self.assertEqual(expected_exec, fake_utils.fake_execute_get_log())
        self.assertEqual(4, mock_update_access.call_count)

    def test_snapshot_update_access(self):
        access_rules = [{
            'access_type': 'ip',
            'access_to': '1.1.1.1',
            'access_level': 'ro',
        }]

        add_rules = [{
            'access_type': 'ip',
            'access_to': '2.2.2.2',
            'access_level': 'ro',
        }]

        delete_rules = [{
            'access_type': 'ip',
            'access_to': '3.3.3.3',
            'access_level': 'ro',
        }]

        self._driver.snapshot_update_access(self._context, self.snapshot,
                                            access_rules, add_rules,
                                            delete_rules)

        (self._driver._helpers[self.snapshot['share']['share_proto']].
            update_access.assert_called_once_with(
            self.server, self.snapshot['name'],
            access_rules, add_rules=add_rules, delete_rules=delete_rules))

    @mock.patch.object(timeutils, 'utcnow', mock.Mock(
                       return_value='fake_date'))
    def test_update_share_usage_size(self):
        mount_path = self._get_mount_path(self.share)
        self._os.path.exists.return_value = True
        self.mock_object(
            self._driver,
            '_execute',
            mock.Mock(return_value=(
                "Mounted on                    Used "
                + mount_path + "               1G", None)))

        update_shares = self._driver.update_share_usage_size(
            self._context, [self.share, ])
        self._os.path.exists.assert_called_with(mount_path)
        self.assertEqual(
            [{'id': 'fakeid', 'used_size': '1',
              'gathered_at': 'fake_date'}],
            update_shares)
        self._driver._execute.assert_called_once_with(
            'df', '-l', '--output=target,used',
            '--block-size=g')

    @mock.patch.object(timeutils, 'utcnow', mock.Mock(
                       return_value='fake_date'))
    def test_update_share_usage_size_multiple_share(self):
        share1 = fake_share(id='fakeid_get_fail', name='get_fail')
        share2 = fake_share(id='fakeid_success', name='get_success')
        share3 = fake_share(id='fakeid_not_exist', name='get_not_exist')

        mount_path2 = self._get_mount_path(share2)
        mount_path3 = self._get_mount_path(share3)
        self._os.path.exists.side_effect = [True, True, False]
        self.mock_object(
            self._driver,
            '_execute',
            mock.Mock(return_value=(
                "Mounted on                    Used "
                + mount_path2 + "               1G", None)))

        update_shares = self._driver.update_share_usage_size(
            self._context, [share1, share2, share3])
        self._os.path.exists.assert_called_with(mount_path3)
        self.assertEqual(
            [{'gathered_at': 'fake_date',
              'id': 'fakeid_success', 'used_size': '1'}],
            update_shares)
        self._driver._execute.assert_called_with(
            'df', '-l', '--output=target,used',
            '--block-size=g')

    def test_update_share_usage_size_fail(self):
        def _fake_exec(*args, **kwargs):
            raise exception.ProcessExecutionError(stderr="error")

        self.mock_object(self._driver, '_execute', _fake_exec)
        self.assertRaises(exception.ProcessExecutionError,
                          self._driver.update_share_usage_size,
                          self._context,
                          [self.share])

    def test_get_backend_info(self):
        backend_info = self._driver.get_backend_info(self._context)

        self.assertEqual(
            {'export_ips': ','.join(self.server['public_addresses']),
             'db_version': mock.ANY},
            backend_info)



# Copyright 2016 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
:mod:`nexenta.jsonrpc` -- Nexenta-specific JSON RPC client
=====================================================================

.. automodule:: nexenta.jsonrpc
"""

import base64
import json
import time

from oslo_log import log
from oslo_serialization import jsonutils
import requests
# pylint: disable=no-member,import-error
from requests.packages.urllib3 import exceptions
requests.packages.urllib3.disable_warnings(exceptions.InsecureRequestWarning)
requests.packages.urllib3.disable_warnings(
    exceptions.InsecurePlatformWarning)
# pylint: enable=no-member,import-error

from manila import exception
from manila.i18n import _


LOG = log.getLogger(__name__)
session = requests.Session()


class NexentaJSONProxy(object):
    def __init__(self, scheme, host, port, user,
                 password, method='get'):
        self.scheme = scheme
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.method = method

    @property
    def url(self):
        return '%s://%s:%s/' % (self.scheme, self.host, self.port)

    def __getattr__(self, method='get'):
        if method:
            return NexentaJSONProxy(
                self.scheme, self.host, self.port,
                self.user, self.password, method)

    def __hash__(self):
        return self.url.__hash__()

    def __repr__(self):
        return 'NEF proxy: %s' % self.url

    def __call__(self, path, data=None):
        auth = base64.b64encode(
            ('%s:%s' % (self.user, self.password)).encode('utf-8'))
        url = self.url + path

        if data:
            data = jsonutils.dumps(data)

        LOG.debug('Sending JSON to url: %s, data: %s, method: %s',
                  path, data, self.method)
        session.headers.update({'Content-Type': 'application/json'})

        response = getattr(session, self.method)(
            url, data=data, verify=False)
        if response.status_code in (401, 403):
            LOG.debug('Login requested by NexentaStor')
            if self.scheme == 'http':
                session.headers.update({'Authorization': 'Basic %s' % auth})
            else:
                session.headers.update(
                    {'Authorization': 'Bearer %s' % self.https_auth()})
            LOG.debug('Re-sending JSON to url: %s, data: %s, method: %s',
                      path, data, self.method)
            response = getattr(session, self.method)(
                url, data=data, verify=False)
        self.check_error(response)
        content = json.loads(response.content) if response.content else None
        LOG.debug("Got response: %(code)s %(reason)s %(content)s", {
            'code': response.status_code,
            'reason': response.reason,
            'content': content})
        response.close()

        if response.status_code == 202 and content:
            url = self.url + content['links'][0]['href']
            keep_going = True
            while keep_going:
                time.sleep(1)
                response = session.get(url, verify=False)
                self.check_error(response)
                LOG.debug("Got response: %(code)s %(reason)s", {
                    'code': response.status_code,
                    'reason': response.reason})
                content = json.loads(
                    response.content) if response.content else None
                keep_going = response.status_code == 202
                response.close()
        return content

    def https_auth(self):
        url = self.url + 'auth/login'
        data = jsonutils.dumps(
            {'username': self.user, 'password': self.password})
        response = session.post(
            url, data=data, verify=False)
        content = json.loads(response.content) if response.content else None
        LOG.debug("Got response: %(code)s %(reason)s %(content)s", {
            'code': response.status_code,
            'reason': response.reason,
            'content': content})
        response.close()
        return content['token']

    def check_error(self, response):
        code = response.status_code
        if code not in (200, 201, 202):
            reason = response.reason
            content = json.loads(
                response.content) if response.content else None
            response.close()
            if content and 'code' in content:
                message = content.get(
                    'message', 'Message is not specified by Nexenta REST')
                raise exception.NexentaException(
                    reason=message, code=content['code'])
            raise exception.NexentaException(
                reason=_(
                    'Got bad response: %(code)s %(reason)s %(content)s') % {
                        'code': code, 'reason': reason, 'content': content})



# Copyright 2015 Clinton Knight
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock
from oslo_serialization import jsonutils
from oslo_utils import encodeutils

from manila.api.openstack import api_version_request
from manila.api.openstack import wsgi
from manila.api.v1 import router
from manila.api import versions
from manila import test
from manila.tests.api import fakes


version_header_name = 'X-OpenStack-Manila-API-Version'
experimental_header_name = 'X-OpenStack-Manila-API-Experimental'


@ddt.ddt
class VersionsControllerTestCase(test.TestCase):

    def setUp(self):
        super(VersionsControllerTestCase, self).setUp()
        self.wsgi_apps = (versions.VersionsRouter(), router.APIRouter())

    @ddt.data('1.0', '1.1', '2.0', '3.0')
    def test_versions_root(self, version):
        req = fakes.HTTPRequest.blank('/', base_url='http://localhost')
        req.method = 'GET'
        req.content_type = 'application/json'
        req.headers = {version_header_name: version}

        response = req.get_response(versions.VersionsRouter())
        self.assertEqual(300, response.status_int)
        body = jsonutils.loads(response.body)
        version_list = body['versions']

        ids = [v['id'] for v in version_list]
        self.assertEqual({'v1.0', 'v2.0'}, set(ids))
        self.assertNotIn(version_header_name, response.headers)
        self.assertNotIn('Vary', response.headers)

        v1 = [v for v in version_list if v['id'] == 'v1.0'][0]
        self.assertEqual('', v1.get('min_version'))
        self.assertEqual('', v1.get('version'))
        self.assertEqual('DEPRECATED', v1.get('status'))

        v2 = [v for v in version_list if v['id'] == 'v2.0'][0]
        self.assertEqual(api_version_request._MIN_API_VERSION,
                         v2.get('min_version'))
        self.assertEqual(api_version_request._MAX_API_VERSION,
                         v2.get('version'))
        self.assertEqual('CURRENT', v2.get('status'))

    @ddt.data('1.0',
              '1.1',
              api_version_request._MIN_API_VERSION,
              api_version_request._MAX_API_VERSION)
    def test_versions_v1(self, version):
        req = fakes.HTTPRequest.blank('/', base_url='http://localhost/v1')
        req.method = 'GET'
        req.content_type = 'application/json'
        req.headers = {version_header_name: version}

        response = req.get_response(router.APIRouter())
        self.assertEqual(200, response.status_int)
        body = jsonutils.loads(response.body)
        version_list = body['versions']

        ids = [v['id'] for v in version_list]
        self.assertEqual({'v1.0'}, set(ids))
        self.assertEqual('1.0', response.headers[version_header_name])
        self.assertEqual(version_header_name, response.headers['Vary'])
        self.assertEqual('', version_list[0].get('min_version'))
        self.assertEqual('', version_list[0].get('version'))
        self.assertEqual('DEPRECATED', version_list[0].get('status'))

    @ddt.data(api_version_request._MIN_API_VERSION,
              api_version_request._MAX_API_VERSION)
    def test_versions_v2(self, version):
        req = fakes.HTTPRequest.blank('/', base_url='http://localhost/v2')
        req.method = 'GET'
        req.content_type = 'application/json'
        req.headers = {version_header_name: version}

        response = req.get_response(router.APIRouter())
        self.assertEqual(200, response.status_int)
        body = jsonutils.loads(response.body)
        version_list = body['versions']

        ids = [v['id'] for v in version_list]
        self.assertEqual({'v2.0'}, set(ids))
        self.assertEqual(version, response.headers[version_header_name])
        self.assertEqual(version_header_name, response.headers['Vary'])

        v2 = [v for v in version_list if v['id'] == 'v2.0'][0]
        self.assertEqual(api_version_request._MIN_API_VERSION,
                         v2.get('min_version'))
        self.assertEqual(api_version_request._MAX_API_VERSION,
                         v2.get('version'))

    def test_versions_version_invalid(self):
        req = fakes.HTTPRequest.blank('/', base_url='http://localhost/v2')
        req.method = 'GET'
        req.content_type = 'application/json'
        req.headers = {version_header_name: '2.0.1'}

        for app in self.wsgi_apps:
            response = req.get_response(app)

            self.assertEqual(400, response.status_int)

    def test_versions_version_not_found(self):
        api_version_request_3_0 = api_version_request.APIVersionRequest('3.0')
        self.mock_object(api_version_request,
                         'max_api_version',
                         mock.Mock(return_value=api_version_request_3_0))

        class Controller(wsgi.Controller):
            @wsgi.Controller.api_version('2.0', '2.0')
            def index(self, req):
                return 'off'

        req = fakes.HTTPRequest.blank('/tests', base_url='http://localhost/v2')
        req.headers = {version_header_name: '2.5'}
        app = fakes.TestRouter(Controller())

        response = req.get_response(app)

        self.assertEqual(404, response.status_int)

    def test_versions_version_not_acceptable(self):
        req = fakes.HTTPRequest.blank('/', base_url='http://localhost/v2')
        req.method = 'GET'
        req.content_type = 'application/json'
        req.headers = {version_header_name: '3.0'}

        response = req.get_response(router.APIRouter())

        self.assertEqual(406, response.status_int)
        self.assertEqual('3.0', response.headers[version_header_name])
        self.assertEqual(version_header_name, response.headers['Vary'])

    @ddt.data(['2.5', 200], ['2.55', 404])
    @ddt.unpack
    def test_req_version_matches(self, version, HTTP_ret):
        version_request = api_version_request.APIVersionRequest(version)
        self.mock_object(api_version_request,
                         'max_api_version',
                         mock.Mock(return_value=version_request))

        class Controller(wsgi.Controller):

            @wsgi.Controller.api_version('2.0', '2.6')
            def index(self, req):
                return 'off'

        req = fakes.HTTPRequest.blank('/tests', base_url='http://localhost/v2')
        req.headers = {version_header_name: version}
        app = fakes.TestRouter(Controller())

        response = req.get_response(app)

        if HTTP_ret == 200:
            self.assertEqual(b'off', response.body)
        elif HTTP_ret == 404:
            self.assertNotEqual(b'off', response.body)
        self.assertEqual(HTTP_ret, response.status_int)

    @ddt.data(['2.5', 'older'], ['2.37', 'newer'])
    @ddt.unpack
    def test_req_version_matches_with_if(self, version, ret_val):
        version_request = api_version_request.APIVersionRequest(version)
        self.mock_object(api_version_request,
                         'max_api_version',
                         mock.Mock(return_value=version_request))

        class Controller(wsgi.Controller):

            def index(self, req):
                req_version = req.api_version_request
                if req_version.matches('2.1', '2.8'):
                    return 'older'
                if req_version.matches('2.9', '2.88'):
                    return 'newer'

        req = fakes.HTTPRequest.blank('/tests', base_url='http://localhost/v2')
        req.headers = {version_header_name: version}
        app = fakes.TestRouter(Controller())

        response = req.get_response(app)

        resp = encodeutils.safe_decode(response.body, incoming='utf-8')
        self.assertEqual(ret_val, resp)
        self.assertEqual(200, response.status_int)

    @ddt.data(['2.5', 'older'], ['2.37', 'newer'])
    @ddt.unpack
    def test_req_version_matches_with_None(self, version, ret_val):
        version_request = api_version_request.APIVersionRequest(version)
        self.mock_object(api_version_request,
                         'max_api_version',
                         mock.Mock(return_value=version_request))

        class Controller(wsgi.Controller):

            def index(self, req):
                req_version = req.api_version_request
                if req_version.matches(None, '2.8'):
                    return 'older'
                if req_version.matches('2.9', None):
                    return 'newer'

        req = fakes.HTTPRequest.blank('/tests', base_url='http://localhost/v2')
        req.headers = {version_header_name: version}
        app = fakes.TestRouter(Controller())

        response = req.get_response(app)

        resp = encodeutils.safe_decode(response.body, incoming='utf-8')
        self.assertEqual(ret_val, resp)
        self.assertEqual(200, response.status_int)

    def test_req_version_matches_with_None_None(self):
        version_request = api_version_request.APIVersionRequest('2.39')
        self.mock_object(api_version_request,
                         'max_api_version',
                         mock.Mock(return_value=version_request))

        class Controller(wsgi.Controller):

            def index(self, req):
                req_version = req.api_version_request
                # This case is artificial, and will return True
                if req_version.matches(None, None):
                    return "Pass"

        req = fakes.HTTPRequest.blank('/tests', base_url='http://localhost/v2')
        req.headers = {version_header_name: '2.39'}
        app = fakes.TestRouter(Controller())

        response = req.get_response(app)

        resp = encodeutils.safe_decode(response.body, incoming='utf-8')
        self.assertEqual("Pass", resp)
        self.assertEqual(200, response.status_int)


@ddt.ddt
class ExperimentalAPITestCase(test.TestCase):

    class Controller(wsgi.Controller):
        @wsgi.Controller.api_version('2.0', '2.0')
        def index(self, req):
            return {'fake_key': 'fake_value'}

        @wsgi.Controller.api_version('2.1', '2.1', experimental=True)  # noqa
        def index(self, req):  # pylint: disable=function-redefined
            return {'fake_key': 'fake_value'}

    def setUp(self):
        super(ExperimentalAPITestCase, self).setUp()
        self.app = fakes.TestRouter(ExperimentalAPITestCase.Controller())
        self.req = fakes.HTTPRequest.blank('/tests',
                                           base_url='http://localhost/v2')

    @ddt.data(True, False)
    def test_stable_api_always_called(self, experimental):

        self.req.headers = {version_header_name: '2.0'}
        if experimental:
            self.req.headers[experimental_header_name] = experimental
        response = self.req.get_response(self.app)

        self.assertEqual(200, response.status_int)
        self.assertEqual('2.0', response.headers[version_header_name])

        if experimental:
            self.assertEqual('%s' % experimental,
                             response.headers.get(experimental_header_name))
        else:
            self.assertNotIn(experimental_header_name, response.headers)

    def test_experimental_api_called_when_requested(self):

        self.req.headers = {
            version_header_name: '2.1',
            experimental_header_name: 'True',
        }
        response = self.req.get_response(self.app)

        self.assertEqual(200, response.status_int)
        self.assertEqual('2.1', response.headers[version_header_name])
        self.assertTrue(response.headers.get(experimental_header_name))

    def test_experimental_api_not_called_when_not_requested(self):

        self.req.headers = {version_header_name: '2.1'}
        response = self.req.get_response(self.app)

        self.assertEqual(404, response.status_int)
        self.assertNotIn(experimental_header_name, response.headers)

    def test_experimental_header_returned_in_exception(self):

        api_version_request_3_0 = api_version_request.APIVersionRequest('3.0')
        self.mock_object(api_version_request,
                         'max_api_version',
                         mock.Mock(return_value=api_version_request_3_0))

        self.req.headers = {
            version_header_name: '2.2',
            experimental_header_name: 'True',
        }
        response = self.req.get_response(self.app)

        self.assertEqual(404, response.status_int)
        self.assertTrue(response.headers.get(experimental_header_name))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock
import six
import webob

import inspect

from manila.api.openstack import api_version_request as api_version
from manila.api.openstack import wsgi
from manila import context
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes


@ddt.ddt
class RequestTest(test.TestCase):
    def test_content_type_missing(self):
        request = wsgi.Request.blank('/tests/123', method='POST')
        request.body = six.b("<body />")
        self.assertIsNone(request.get_content_type())

    def test_content_type_unsupported(self):
        request = wsgi.Request.blank('/tests/123', method='POST')
        request.headers["Content-Type"] = "text/html"
        request.body = six.b("asdf<br />")
        self.assertRaises(exception.InvalidContentType,
                          request.get_content_type)

    def test_content_type_with_charset(self):
        request = wsgi.Request.blank('/tests/123')
        request.headers["Content-Type"] = "application/json; charset=UTF-8"
        result = request.get_content_type()
        self.assertEqual("application/json", result)

    def test_content_type_from_accept(self):
        content_type = 'application/json'
        request = wsgi.Request.blank('/tests/123')
        request.headers["Accept"] = content_type

        result = request.best_match_content_type()

        self.assertEqual(content_type, result)

    def test_content_type_from_accept_best(self):
        request = wsgi.Request.blank('/tests/123')
        request.headers["Accept"] = "application/xml, application/json"
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

        request = wsgi.Request.blank('/tests/123')
        request.headers["Accept"] = ("application/json; q=0.3, "
                                     "application/xml; q=0.9")
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

    def test_content_type_from_query_extension(self):
        request = wsgi.Request.blank('/tests/123.json')
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

        request = wsgi.Request.blank('/tests/123.invalid')
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

    def test_content_type_accept_default(self):
        request = wsgi.Request.blank('/tests/123.unsupported')
        request.headers["Accept"] = "application/unsupported1"
        result = request.best_match_content_type()
        self.assertEqual("application/json", result)

    def test_cache_and_retrieve_resources(self):
        request = wsgi.Request.blank('/foo')
        # Test that trying to retrieve a cached object on
        # an empty cache fails gracefully
        self.assertIsNone(request.cached_resource())
        self.assertIsNone(request.cached_resource_by_id('r-0'))

        resources = [{'id': 'r-%s' % x} for x in range(3)]

        # Cache an empty list of resources using the default name
        request.cache_resource([])
        self.assertEqual({}, request.cached_resource())
        self.assertIsNone(request.cached_resource('r-0'))
        # Cache some resources
        request.cache_resource(resources[:2])
        # Cache  one resource
        request.cache_resource(resources[2])
        # Cache  a different resource name
        other_resource = {'id': 'o-0'}
        request.cache_resource(other_resource, name='other-resource')

        self.assertEqual(resources[0], request.cached_resource_by_id('r-0'))
        self.assertEqual(resources[1], request.cached_resource_by_id('r-1'))
        self.assertEqual(resources[2], request.cached_resource_by_id('r-2'))
        self.assertIsNone(request.cached_resource_by_id('r-3'))
        self.assertEqual(
            {'r-0': resources[0], 'r-1': resources[1], 'r-2': resources[2]},
            request.cached_resource())
        self.assertEqual(
            other_resource,
            request.cached_resource_by_id('o-0', name='other-resource'))

    @ddt.data(
        'share_type',
    )
    def test_cache_and_retrieve_resources_by_resource(self, resource_name):
        cache_all_func = 'cache_db_%ss' % resource_name
        cache_one_func = 'cache_db_%s' % resource_name
        get_db_all_func = 'get_db_%ss' % resource_name
        get_db_one_func = 'get_db_%s' % resource_name

        r = wsgi.Request.blank('/foo')
        amount = 5
        res_range = range(amount)
        resources = [{'id': 'id%s' % x} for x in res_range]

        # Store 2
        getattr(r, cache_all_func)(resources[:amount - 1])
        # Store 1
        getattr(r, cache_one_func)(resources[amount - 1])

        for i in res_range:
            self.assertEqual(
                resources[i],
                getattr(r, get_db_one_func)('id%s' % i),
            )
        self.assertIsNone(getattr(r, get_db_one_func)('id%s' % amount))
        self.assertEqual(
            {'id%s' % i: resources[i] for i in res_range},
            getattr(r, get_db_all_func)())

    def test_set_api_version_request_exception(self):
        min_version = api_version.APIVersionRequest('2.0')
        max_version = api_version.APIVersionRequest('2.45')
        self.mock_object(api_version, 'max_api_version',
                         mock.Mock(return_value=max_version))
        self.mock_object(api_version, 'min_api_version',
                         mock.Mock(return_value=min_version))
        headers = {'X-OpenStack-Manila-API-Version': '2.50'}
        request = wsgi.Request.blank(
            'https://openstack.acme.com/v2/shares', method='GET',
            headers=headers, script_name='/v2/shares')

        self.assertRaises(exception.InvalidGlobalAPIVersion,
                          request.set_api_version_request)
        self.assertEqual(api_version.APIVersionRequest('2.50'),
                         request.api_version_request)

    @ddt.data('', '/share', '/v1', '/v2/shares', '/v1.1/', '/share/v1',
              '/shared-file-sytems/v2', '/share/v3.5/share-replicas',
              '/shared-file-sytems/v2/shares/xyzzy/action')
    def test_set_api_version_request(self, resource):
        min_version = api_version.APIVersionRequest('2.0')
        max_version = api_version.APIVersionRequest('3.0')
        self.mock_object(api_version, 'max_api_version',
                         mock.Mock(return_value=max_version))
        self.mock_object(api_version, 'min_api_version',
                         mock.Mock(return_value=min_version))
        request = wsgi.Request.blank(
            'https://openstack.acme.com%s' % resource, method='GET',
            headers={'X-OpenStack-Manila-API-Version': '2.117'},
            script_name=resource)

        self.assertIsNone(request.set_api_version_request())

        if not resource or not ('/v1' in resource or '/v2' in resource):
            self.assertEqual(api_version.APIVersionRequest(),
                             request.api_version_request)
        elif 'v1' in resource:
            self.assertEqual(api_version.APIVersionRequest('1.0'),
                             request.api_version_request)
        else:
            self.assertEqual(api_version.APIVersionRequest('2.117'),
                             request.api_version_request)

    def test_set_api_version_request_no_version_header(self):
        min_version = api_version.APIVersionRequest('2.0')
        max_version = api_version.APIVersionRequest('2.45')
        self.mock_object(api_version, 'max_api_version',
                         mock.Mock(return_value=max_version))
        self.mock_object(api_version, 'min_api_version',
                         mock.Mock(return_value=min_version))
        headers = {}
        request = wsgi.Request.blank(
            'https://openstack.acme.com/v2/shares', method='GET',
            headers=headers, script_name='/v2/shares')

        self.assertIsNone(request.set_api_version_request())

        self.assertEqual(api_version.APIVersionRequest('2.0'),
                         request.api_version_request)

    @ddt.data(None, 'true', 'false')
    def test_set_api_version_request_experimental_header(self, experimental):
        min_version = api_version.APIVersionRequest('2.0')
        max_version = api_version.APIVersionRequest('2.45')
        self.mock_object(api_version, 'max_api_version',
                         mock.Mock(return_value=max_version))
        self.mock_object(api_version, 'min_api_version',
                         mock.Mock(return_value=min_version))
        headers = {'X-OpenStack-Manila-API-Version': '2.38'}
        if experimental:
            headers['X-OpenStack-Manila-API-Experimental'] = experimental
        request = wsgi.Request.blank(
            'https://openstack.acme.com/v2/shares', method='GET',
            headers=headers, script_name='/v2/shares')

        self.assertIsNone(request.set_api_version_request())

        self.assertEqual(request.api_version_request,
                         api_version.APIVersionRequest('2.38'))

        expected_experimental = experimental == 'true' or False
        self.assertEqual(expected_experimental,
                         request.api_version_request.experimental)


class ActionDispatcherTest(test.TestCase):
    def test_dispatch(self):
        serializer = wsgi.ActionDispatcher()
        serializer.create = lambda x: 'pants'
        self.assertEqual('pants', serializer.dispatch({}, action='create'))

    def test_dispatch_action_None(self):
        serializer = wsgi.ActionDispatcher()
        serializer.create = lambda x: 'pants'
        serializer.default = lambda x: 'trousers'
        self.assertEqual('trousers', serializer.dispatch({}, action=None))

    def test_dispatch_default(self):
        serializer = wsgi.ActionDispatcher()
        serializer.create = lambda x: 'pants'
        serializer.default = lambda x: 'trousers'
        self.assertEqual('trousers', serializer.dispatch({}, action='update'))


class DictSerializerTest(test.TestCase):
    def test_dispatch_default(self):
        serializer = wsgi.DictSerializer()
        self.assertEqual('', serializer.serialize({}, 'update'))


class JSONDictSerializerTest(test.TestCase):
    def test_json(self):
        input_dict = dict(servers=dict(a=(2, 3)))
        expected_json = six.b('{"servers":{"a":[2,3]}}')
        serializer = wsgi.JSONDictSerializer()
        result = serializer.serialize(input_dict)
        result = result.replace(six.b('\n'),
                                six.b('')).replace(six.b(' '), six.b(''))
        self.assertEqual(expected_json, result)


class TextDeserializerTest(test.TestCase):
    def test_dispatch_default(self):
        deserializer = wsgi.TextDeserializer()
        self.assertEqual({}, deserializer.deserialize({}, 'update'))


class JSONDeserializerTest(test.TestCase):
    def test_json(self):
        data = """{"a": {
                "a1": "1",
                "a2": "2",
                "bs": ["1", "2", "3", {"c": {"c1": "1"}}],
                "d": {"e": "1"},
                "f": "1"}}"""
        as_dict = {
            'body': {
                'a': {
                    'a1': '1',
                    'a2': '2',
                    'bs': ['1', '2', '3', {'c': {'c1': '1'}}],
                    'd': {'e': '1'},
                    'f': '1',
                },
            },
        }
        deserializer = wsgi.JSONDeserializer()
        self.assertEqual(as_dict, deserializer.deserialize(data))


class ResourceTest(test.TestCase):
    def test_resource_call(self):
        class Controller(object):
            def index(self, req):
                return 'off'

        req = webob.Request.blank('/tests')
        app = fakes.TestRouter(Controller())
        response = req.get_response(app)
        self.assertEqual(six.b('off'), response.body)
        self.assertEqual(200, response.status_int)

    def test_resource_not_authorized(self):
        class Controller(object):
            def index(self, req):
                raise exception.NotAuthorized()

        req = webob.Request.blank('/tests')
        app = fakes.TestRouter(Controller())
        response = req.get_response(app)
        self.assertEqual(403, response.status_int)

    def test_dispatch(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)
        method, extensions = resource.get_method(None, 'index', None, '')
        actual = resource.dispatch(method, None, {'pants': 'off'})
        expected = 'off'
        self.assertEqual(expected, actual)

    def test_get_method_undefined_controller_action(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertRaises(AttributeError, resource.get_method,
                          None, 'create', None, '')

    def test_get_method_action_json(self):
        class Controller(wsgi.Controller):
            @wsgi.action('fooAction')
            def _action_foo(self, req, id, body):
                return body

        controller = Controller()
        resource = wsgi.Resource(controller)
        method, extensions = resource.get_method(None, 'action',
                                                 'application/json',
                                                 '{"fooAction": true}')
        self.assertEqual(controller._action_foo, method)

    def test_get_method_action_bad_body(self):
        class Controller(wsgi.Controller):
            @wsgi.action('fooAction')
            def _action_foo(self, req, id, body):
                return body

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertRaises(exception.MalformedRequestBody, resource.get_method,
                          None, 'action', 'application/json', '{}')

    def test_get_method_unknown_controller_action(self):
        class Controller(wsgi.Controller):
            @wsgi.action('fooAction')
            def _action_foo(self, req, id, body):
                return body

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertRaises(KeyError, resource.get_method,
                          None, 'action', 'application/json',
                          '{"barAction": true}')

    def test_get_method_action_method(self):
        class Controller(object):
            def action(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)
        method, extensions = resource.get_method(None, 'action',
                                                 'application/xml',
                                                 '<fooAction>true</fooAction')
        self.assertEqual(controller.action, method)

    def test_get_action_args(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        env = {
            'wsgiorg.routing_args': [None, {
                'controller': None,
                'format': None,
                'action': 'update',
                'id': 12,
            }],
        }

        expected = {'action': 'update', 'id': 12}

        self.assertEqual(expected, resource.get_action_args(env))

    def test_get_body_bad_content(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        request = wsgi.Request.blank('/', method='POST')
        request.headers['Content-Type'] = 'application/none'
        request.body = six.b('foo')

        content_type, body = resource.get_body(request)
        self.assertIsNone(content_type)
        self.assertEqual('', body)

    def test_get_body_no_content_type(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        request = wsgi.Request.blank('/', method='POST')
        request.body = six.b('foo')

        content_type, body = resource.get_body(request)
        self.assertIsNone(content_type)
        self.assertEqual('', body)

    def test_get_body_no_content_body(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        request = wsgi.Request.blank('/', method='POST')
        request.headers['Content-Type'] = 'application/json'
        request.body = six.b('')

        content_type, body = resource.get_body(request)
        self.assertIsNone(content_type)
        self.assertEqual('', body)

    def test_get_body(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        request = wsgi.Request.blank('/', method='POST')
        request.headers['Content-Type'] = 'application/json'
        request.body = six.b('foo')

        content_type, body = resource.get_body(request)
        self.assertEqual('application/json', content_type)
        self.assertEqual(six.b('foo'), body)

    def test_deserialize_badtype(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertRaises(exception.InvalidContentType,
                          resource.deserialize,
                          controller.index, 'application/none', 'foo')

    def test_deserialize_default(self):
        class JSONDeserializer(object):
            def deserialize(self, body):
                return 'json'

        class XMLDeserializer(object):
            def deserialize(self, body):
                return 'xml'

        class Controller(object):
            @wsgi.deserializers(xml=XMLDeserializer)
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller, json=JSONDeserializer)

        obj = resource.deserialize(controller.index, 'application/json', 'foo')
        self.assertEqual('json', obj)

    def test_deserialize_decorator(self):
        class JSONDeserializer(object):
            def deserialize(self, body):
                return 'json'

        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller, json=JSONDeserializer)

        obj = resource.deserialize(controller.index, 'application/json', 'foo')
        self.assertEqual('json', obj)

    def test_register_actions(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        class ControllerExtended(wsgi.Controller):
            @wsgi.action('fooAction')
            def _action_foo(self, req, id, body):
                return body

            @wsgi.action('barAction')
            def _action_bar(self, req, id, body):
                return body

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertEqual({}, resource.wsgi_actions)
        extended = ControllerExtended()
        resource.register_actions(extended)
        self.assertEqual({'fooAction': extended._action_foo,
                          'barAction': extended._action_bar, },
                         resource.wsgi_actions)

    def test_register_extensions(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        class ControllerExtended(wsgi.Controller):
            @wsgi.extends
            def index(self, req, resp_obj, pants=None):
                return None

            @wsgi.extends(action='fooAction')
            def _action_foo(self, req, resp, id, body):
                return None

        controller = Controller()
        resource = wsgi.Resource(controller)
        self.assertEqual({}, resource.wsgi_extensions)
        self.assertEqual({}, resource.wsgi_action_extensions)

        extended = ControllerExtended()
        resource.register_extensions(extended)
        self.assertEqual({'index': [extended.index]}, resource.wsgi_extensions)
        self.assertEqual({'fooAction': [extended._action_foo]},
                         resource.wsgi_action_extensions)

    def test_get_method_extensions(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        class ControllerExtended(wsgi.Controller):
            @wsgi.extends
            def index(self, req, resp_obj, pants=None):
                return None

        controller = Controller()
        extended = ControllerExtended()
        resource = wsgi.Resource(controller)
        resource.register_extensions(extended)
        method, extensions = resource.get_method(None, 'index', None, '')
        self.assertEqual(controller.index, method)
        self.assertEqual([extended.index], extensions)

    def test_get_method_action_extensions(self):
        class Controller(wsgi.Controller):
            def index(self, req, pants=None):
                return pants

            @wsgi.action('fooAction')
            def _action_foo(self, req, id, body):
                return body

        class ControllerExtended(wsgi.Controller):
            @wsgi.extends(action='fooAction')
            def _action_foo(self, req, resp_obj, id, body):
                return None

        controller = Controller()
        extended = ControllerExtended()
        resource = wsgi.Resource(controller)
        resource.register_extensions(extended)
        method, extensions = resource.get_method(None, 'action',
                                                 'application/json',
                                                 '{"fooAction": true}')
        self.assertEqual(controller._action_foo, method)
        self.assertEqual([extended._action_foo], extensions)

    def test_get_method_action_whitelist_extensions(self):
        class Controller(wsgi.Controller):
            def index(self, req, pants=None):
                return pants

        class ControllerExtended(wsgi.Controller):
            @wsgi.action('create')
            def _create(self, req, body):
                pass

            @wsgi.action('delete')
            def _delete(self, req, id):
                pass

        controller = Controller()
        extended = ControllerExtended()
        resource = wsgi.Resource(controller)
        resource.register_actions(extended)

        method, extensions = resource.get_method(None, 'create',
                                                 'application/json',
                                                 '{"create": true}')
        self.assertEqual(extended._create, method)
        self.assertEqual([], extensions)

        method, extensions = resource.get_method(None, 'delete', None, None)
        self.assertEqual(extended._delete, method)
        self.assertEqual([], extensions)

    def test_pre_process_extensions_regular(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req, resp_obj):
            called.append(1)
            return None

        def extension2(req, resp_obj):
            called.append(2)
            return None

        extensions = [extension1, extension2]
        response, post = resource.pre_process_extensions(extensions, None, {})
        self.assertEqual([], called)
        self.assertIsNone(response)
        self.assertEqual([extension2, extension1], list(post))

    def test_pre_process_extensions_generator(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req):
            called.append('pre1')
            resp_obj = yield
            self.assertIsNone(resp_obj)
            called.append('post1')

        def extension2(req):
            called.append('pre2')
            resp_obj = yield
            self.assertIsNone(resp_obj)
            called.append('post2')

        extensions = [extension1, extension2]
        response, post = resource.pre_process_extensions(extensions, None, {})
        post = list(post)
        self.assertEqual(['pre1', 'pre2'], called)
        self.assertIsNone(response)
        self.assertEqual(2, len(post))
        self.assertTrue(inspect.isgenerator(post[0]))
        self.assertTrue(inspect.isgenerator(post[1]))

        for gen in post:
            try:
                gen.send(None)
            except StopIteration:
                continue

        self.assertEqual(['pre1', 'pre2', 'post2', 'post1'], called)

    def test_pre_process_extensions_generator_response(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req):
            called.append('pre1')
            yield 'foo'

        def extension2(req):
            called.append('pre2')

        extensions = [extension1, extension2]
        response, post = resource.pre_process_extensions(extensions, None, {})
        self.assertEqual(['pre1'], called)
        self.assertEqual('foo', response)
        self.assertEqual([], post)

    def test_post_process_extensions_regular(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req, resp_obj):
            called.append(1)
            return None

        def extension2(req, resp_obj):
            called.append(2)
            return None

        response = resource.post_process_extensions([extension2, extension1],
                                                    None, None, {})
        self.assertEqual([2, 1], called)
        self.assertIsNone(response)

    def test_post_process_extensions_regular_response(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req, resp_obj):
            called.append(1)
            return None

        def extension2(req, resp_obj):
            called.append(2)
            return 'foo'

        response = resource.post_process_extensions([extension2, extension1],
                                                    None, None, {})
        self.assertEqual([2], called)
        self.assertEqual('foo', response)

    def test_post_process_extensions_version_not_found(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req, resp_obj):
            called.append(1)
            return 'bar'

        def extension2(req, resp_obj):
            raise exception.VersionNotFoundForAPIMethod(version='fake_version')

        response = resource.post_process_extensions([extension2, extension1],
                                                    None, None, {})
        self.assertEqual([1], called)
        self.assertEqual('bar', response)

    def test_post_process_extensions_generator(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req):
            resp_obj = yield
            self.assertIsNone(resp_obj)
            called.append(1)

        def extension2(req):
            resp_obj = yield
            self.assertIsNone(resp_obj)
            called.append(2)

        ext1 = extension1(None)
        next(ext1)
        ext2 = extension2(None)
        next(ext2)

        response = resource.post_process_extensions([ext2, ext1],
                                                    None, None, {})

        self.assertEqual([2, 1], called)
        self.assertIsNone(response)

    def test_post_process_extensions_generator_response(self):
        class Controller(object):
            def index(self, req, pants=None):
                return pants

        controller = Controller()
        resource = wsgi.Resource(controller)

        called = []

        def extension1(req):
            resp_obj = yield
            self.assertIsNone(resp_obj)
            called.append(1)

        def extension2(req):
            resp_obj = yield
            self.assertIsNone(resp_obj)
            called.append(2)
            yield 'foo'

        ext1 = extension1(None)
        next(ext1)
        ext2 = extension2(None)
        next(ext2)

        response = resource.post_process_extensions([ext2, ext1],
                                                    None, None, {})

        self.assertEqual([2], called)
        self.assertEqual('foo', response)


class ResponseObjectTest(test.TestCase):
    def test_default_code(self):
        robj = wsgi.ResponseObject({})
        self.assertEqual(200, robj.code)

    def test_modified_code(self):
        robj = wsgi.ResponseObject({})
        robj._default_code = 202
        self.assertEqual(202, robj.code)

    def test_override_default_code(self):
        robj = wsgi.ResponseObject({}, code=404)
        self.assertEqual(404, robj.code)

    def test_override_modified_code(self):
        robj = wsgi.ResponseObject({}, code=404)
        robj._default_code = 202
        self.assertEqual(404, robj.code)

    def test_set_header(self):
        robj = wsgi.ResponseObject({})
        robj['Header'] = 'foo'
        self.assertEqual({'header': 'foo'}, robj.headers)

    def test_get_header(self):
        robj = wsgi.ResponseObject({})
        robj['Header'] = 'foo'
        self.assertEqual('foo', robj['hEADER'])

    def test_del_header(self):
        robj = wsgi.ResponseObject({})
        robj['Header'] = 'foo'
        del robj['hEADER']
        self.assertNotIn('header', robj.headers)

    def test_header_isolation(self):
        robj = wsgi.ResponseObject({})
        robj['Header'] = 'foo'
        hdrs = robj.headers
        hdrs['hEADER'] = 'bar'
        self.assertEqual('foo', robj['hEADER'])

    def test_default_serializers(self):
        robj = wsgi.ResponseObject({})
        self.assertEqual({}, robj.serializers)

    def test_bind_serializers(self):
        robj = wsgi.ResponseObject({}, json='foo')
        robj._bind_method_serializers(dict(xml='bar', json='baz'))
        self.assertEqual(dict(xml='bar', json='foo'), robj.serializers)

    def test_get_serializer(self):
        robj = wsgi.ResponseObject({}, json='json', xml='xml', atom='atom')
        for content_type, mtype in wsgi._MEDIA_TYPE_MAP.items():
            _mtype, serializer = robj.get_serializer(content_type)
            self.assertEqual(mtype, serializer)

    def test_get_serializer_defaults(self):
        robj = wsgi.ResponseObject({})
        default_serializers = dict(json='json', xml='xml', atom='atom')
        for content_type, mtype in wsgi._MEDIA_TYPE_MAP.items():
            self.assertRaises(exception.InvalidContentType,
                              robj.get_serializer, content_type)
            _mtype, serializer = robj.get_serializer(content_type,
                                                     default_serializers)
            self.assertEqual(mtype, serializer)

    def test_serialize(self):
        class JSONSerializer(object):
            def serialize(self, obj):
                return six.b('json')

        class XMLSerializer(object):
            def serialize(self, obj):
                return six.b('xml')

        class AtomSerializer(object):
            def serialize(self, obj):
                return six.b('atom')

        robj = wsgi.ResponseObject({}, code=202,
                                   json=JSONSerializer,
                                   xml=XMLSerializer,
                                   atom=AtomSerializer)
        robj['X-header1'] = 'header1'
        robj['X-header2'] = 'header2'

        for content_type, mtype in wsgi._MEDIA_TYPE_MAP.items():
            request = wsgi.Request.blank('/tests/123')
            response = robj.serialize(request, content_type)

            self.assertEqual(content_type, response.headers['Content-Type'])
            self.assertEqual('header1', response.headers['X-header1'])
            self.assertEqual('header2', response.headers['X-header2'])
            self.assertEqual(202, response.status_int)
            self.assertEqual(six.b(mtype), response.body)


class ValidBodyTest(test.TestCase):

    def setUp(self):
        super(ValidBodyTest, self).setUp()
        self.controller = wsgi.Controller()

    def test_is_valid_body(self):
        body = {'foo': {}}
        self.assertTrue(self.controller.is_valid_body(body, 'foo'))

    def test_is_valid_body_none(self):
        self.assertFalse(self.controller.is_valid_body(None, 'foo'))

    def test_is_valid_body_empty(self):
        self.assertFalse(self.controller.is_valid_body({}, 'foo'))

    def test_is_valid_body_no_entity(self):
        body = {'bar': {}}
        self.assertFalse(self.controller.is_valid_body(body, 'foo'))

    def test_is_valid_body_malformed_entity(self):
        body = {'foo': 'bar'}
        self.assertFalse(self.controller.is_valid_body(body, 'foo'))


class AuthorizeDecoratorTest(test.TestCase):
    class FakeController(wsgi.Controller):
        resource_name = 'fake_resource_name'

        @wsgi.Controller.authorize
        def fake_action_1(self, req):
            pass

        @wsgi.Controller.authorize('fake_action')
        def fake_action_2(self, req):
            pass

    def setUp(self):
        super(AuthorizeDecoratorTest, self).setUp()
        self.controller = self.FakeController()
        self.admin_context = context.get_admin_context()
        self.user_context = fakes.FakeRequestContext
        self.mock_policy_check = self.mock_object(policy, 'check_policy')

    def test_authorize_decorator_no_args(self):
        req = fakes.HTTPRequest.blank('/v2/fake/fake_id')
        req.environ['manila.context'] = self.user_context

        self.controller.fake_action_1(req)

        self.mock_policy_check.assert_called_once_with(
            self.user_context, self.controller.resource_name, 'fake_action_1')

    def test_authorize_decorator_action_name(self):
        req = fakes.HTTPRequest.blank('/v2/fake/fake_id')
        req.environ['manila.context'] = self.admin_context

        self.controller.fake_action_2(req)

        self.mock_policy_check.assert_called_once_with(
            self.admin_context, self.controller.resource_name, 'fake_action')

    def test_authorize_exception(self):
        req = fakes.HTTPRequest.blank('/v2/fake/fake_id')
        req.environ['manila.context'] = self.admin_context
        exc = exception.PolicyNotAuthorized(action='fake_action')

        with mock.patch.object(policy, 'check_policy',
                               mock.Mock(side_effect=exc)):
            self.assertRaises(webob.exc.HTTPForbidden,
                              self.controller.fake_action_2, req)



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add share_export_locations table

Revision ID: 56cdbe267881
Revises: 17115072e1c3
Create Date: 2015-02-27 14:06:30.464315

"""

# revision identifiers, used by Alembic.
revision = '56cdbe267881'
down_revision = '30cb96d995fa'

from alembic import op
import sqlalchemy as sa
from sqlalchemy import func
from sqlalchemy.sql import table


def upgrade():
    export_locations_table = op.create_table(
        'share_export_locations',
        sa.Column('id', sa.Integer, primary_key=True, nullable=False),
        sa.Column('created_at', sa.DateTime),
        sa.Column('updated_at', sa.DateTime),
        sa.Column('deleted_at', sa.DateTime),
        sa.Column('deleted', sa.Integer, default=0),
        sa.Column('path', sa.String(2000)),
        sa.Column('share_id', sa.String(36),
                  sa.ForeignKey('shares.id', name="sel_id_fk")),
        mysql_engine='InnoDB',
        mysql_charset='utf8')

    shares_table = table(
        'shares',
        sa.Column('created_at', sa.DateTime),
        sa.Column('deleted_at', sa.DateTime),
        sa.Column('deleted', sa.Integer),
        sa.Column('export_location', sa.String(length=255)),
        sa.Column('id', sa.String(length=36)),
        sa.Column('updated_at', sa.DateTime))

    export_locations = []
    session = sa.orm.Session(bind=op.get_bind().connect())
    for share in session.query(shares_table).all():
        deleted = share.deleted if isinstance(share.deleted, int) else 0
        export_locations.append({
            'created_at': share.created_at,
            'updated_at': share.updated_at,
            'deleted_at': share.deleted_at,
            'deleted': deleted,
            'share_id': share.id,
            'path': share.export_location,
        })
    op.bulk_insert(export_locations_table, export_locations)

    op.drop_column('shares', 'export_location')
    session.close_all()


def downgrade():
    """Remove share_export_locations table.

    This method can lead to data loss because only first export_location
    is saved in shares table.
    """

    op.add_column('shares',
                  sa.Column('export_location', sa.String(255)))

    export_locations_table = table(
        'share_export_locations',
        sa.Column('share_id', sa.String(length=36)),
        sa.Column('path', sa.String(length=255)),
        sa.Column('updated_at', sa.DateTime),
        sa.Column('deleted', sa.Integer))

    connection = op.get_bind()
    session = sa.orm.Session(bind=connection.connect())
    export_locations = session.query(
        func.min(export_locations_table.c.updated_at),
        export_locations_table.c.share_id,
        export_locations_table.c.path).filter(
            export_locations_table.c.deleted == 0).group_by(
                export_locations_table.c.share_id,
                export_locations_table.c.path).all()

    shares = sa.Table('shares', sa.MetaData(),
                      autoload=True, autoload_with=connection)

    for location in export_locations:
        # pylint: disable=no-value-for-parameter
        update = (shares.update().where(shares.c.id == location.share_id).
                  values(export_location=location.path))
        connection.execute(update)

    op.drop_table('share_export_locations')
    session.close_all()
